import logging

from mpfmc.tests.MpfMcTestCase import MpfMcTestCase
from unittest.mock import MagicMock, call, ANY
from mpfmc.config_collections.playlist import PlaylistInstance

try:
    from mpfmc.core.audio import SoundSystem
    from mpfmc.assets.sound import SoundStealingMethod
except ImportError:
    SoundSystem = None
    SoundStealingMethod = None
    logging.warning("mpfmc.core.audio library could not be loaded. Audio "
                    "features will not be available")


class TestAudioPlaylist(MpfMcTestCase):
    """
    """
    def get_machine_path(self):
        return 'tests/machine_files/audio'

    def get_config_file(self):
        return 'test_audio_playlist.yaml'

    def test_playlist_track(self):
        """ Tests the Playlist track type and its associated assets"""

        if SoundSystem is None or self.mc.sound_system is None:
            log = logging.getLogger('TestAudioPlaylist')
            log.warning("Sound system is not enabled - skipping audio tests")
            self.skipTest("Sound system is not enabled")

        self.assertIsNotNone(self.mc.sound_system)
        interface = self.mc.sound_system.audio_interface
        if interface is None:
            log = logging.getLogger('TestAudioPlaylist')
            log.warning("Sound system audio interface could not be loaded - skipping audio tests")
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(interface)

        # Check playlist track
        track_playlist = interface.get_track_by_name("playlist")
        self.assertIsNotNone(track_playlist)
        self.assertEqual("playlist", track_playlist.name)
        self.assertAlmostEqual(0.6, track_playlist.volume, 1)

        playlist_controller = interface.get_playlist_controller("playlist")
        self.assertIsNotNone(playlist_controller)
        self.assertEqual("playlist", playlist_controller.name)
        self.assertEqual(2.0, playlist_controller.crossfade_time)

        # /sounds/playlist
        self.assertTrue(hasattr(self.mc, 'sounds'))
        self.assertIn('drumbeat_7', self.mc.sounds)
        self.assertIn('hippie_ahead', self.mc.sounds)
        self.assertIn('rainbow_disco_bears', self.mc.sounds)
        self.assertIn('dirty_grinding_beat_loop', self.mc.sounds)

        # Playlists
        self.assertTrue(hasattr(self.mc, 'playlists'))
        self.assertIn('attract_music', self.mc.playlists)
        self.assertListEqual(['drumbeat_7', 'rainbow_disco_bears', 'dirty_grinding_beat_loop', 'hippie_ahead'],
                             self.mc.playlists['attract_music']['sounds'])

        # Create a PlaylistInstance to manipulate directly for testing
        attract_music_playlist = PlaylistInstance('attract_music',
                                                  self.mc.playlists['attract_music'],
                                                  playlist_controller.crossfade_time,
                                                  settings={
                                                      "crossfade_mode": "override",
                                                      "crossfade_time": 1.5
                                                  })
        self.assertIsNotNone(attract_music_playlist)
        self.assertFalse(attract_music_playlist.repeat)
        self.assertEqual(1.5, attract_music_playlist.crossfade_time)

        # Advance through the playlist
        current_sound = attract_music_playlist.get_next_sound_name()
        self.assertEqual('drumbeat_7', current_sound)
        self.assertFalse(attract_music_playlist.end_of_playlist)
        current_sound = attract_music_playlist.get_next_sound_name()
        self.assertEqual('rainbow_disco_bears', current_sound)
        self.assertFalse(attract_music_playlist.end_of_playlist)
        current_sound = attract_music_playlist.get_next_sound_name()
        self.assertEqual('dirty_grinding_beat_loop', current_sound)
        self.assertFalse(attract_music_playlist.end_of_playlist)
        current_sound = attract_music_playlist.get_next_sound_name()
        self.assertEqual('hippie_ahead', current_sound)
        self.assertTrue(attract_music_playlist.end_of_playlist)
        current_sound = attract_music_playlist.get_next_sound_name()
        self.assertIsNone(current_sound)

        # Create another PlaylistInstance, this time enable repeat
        attract_music_playlist = PlaylistInstance('attract_music',
                                                  self.mc.playlists['attract_music'],
                                                  playlist_controller.crossfade_time,
                                                  settings={'repeat': True})
        self.assertIsNotNone(attract_music_playlist)
        self.assertTrue(attract_music_playlist.repeat)

        # Advance through the playlist
        current_sound = attract_music_playlist.get_next_sound_name()
        self.assertEqual('drumbeat_7', current_sound)
        self.assertFalse(attract_music_playlist.end_of_playlist)
        current_sound = attract_music_playlist.get_next_sound_name()
        self.assertEqual('rainbow_disco_bears', current_sound)
        self.assertFalse(attract_music_playlist.end_of_playlist)
        current_sound = attract_music_playlist.get_next_sound_name()
        self.assertEqual('dirty_grinding_beat_loop', current_sound)
        self.assertFalse(attract_music_playlist.end_of_playlist)
        current_sound = attract_music_playlist.get_next_sound_name()
        self.assertEqual('hippie_ahead', current_sound)
        self.assertTrue(attract_music_playlist.end_of_playlist)
        current_sound = attract_music_playlist.get_next_sound_name()
        self.assertEqual('drumbeat_7', current_sound)
        self.assertFalse(attract_music_playlist.end_of_playlist)

        # Mock BCP send method
        self.mc.bcp_processor.send = MagicMock()
        self.mc.bcp_processor.enabled = True

        # Test playlist_player
        self.advance_time()
        self.mc.events.post('play_attract_music')
        self.advance_real_time(1)

        self.mc.bcp_processor.send.assert_has_calls([call('trigger', name='attract_music_played'),
                                                     call('trigger', name='attract_music_sound_changed'),
                                                     call('trigger', sound_instance=ANY, name='drumbeat_7_played')])

        status = track_playlist.get_status()
        self.assertEqual("playing", status[0]['status'])
        self.assertEqual(self.mc.sounds["drumbeat_7"].id, status[0]['sound_id'])
        self.assertEqual("idle", status[1]['status'])
        self.advance_real_time(2)

        self.mc.bcp_processor.send.reset_mock()
        self.mc.events.post('advance_playlist')
        self.advance_real_time(0.25)

        self.mc.bcp_processor.send.assert_has_calls([call('trigger', name='attract_music_sound_changed'),
                                                     call('trigger', sound_instance=ANY,
                                                          name='rainbow_disco_bears_played')])

        status = track_playlist.get_status()
        self.assertEqual("stopping", status[0]['status'])
        self.assertEqual(self.mc.sounds["drumbeat_7"].id, status[0]['sound_id'])
        self.assertEqual("playing", status[1]['status'])
        self.assertEqual(self.mc.sounds["rainbow_disco_bears"].id, status[1]['sound_id'])

        self.mc.bcp_processor.send.reset_mock()
        self.advance_real_time(2)

        self.mc.bcp_processor.send.assert_has_calls([call('trigger', sound_instance=ANY,
                                                          name='drumbeat_7_stopped'),
                                                     call('trigger', sound_instance=ANY,
                                                          name='playlist_playlist_sound_stopped'),
                                                     call('trigger', name='attract_music_sound_stopped')])

        status = track_playlist.get_status()
        self.assertEqual("idle", status[0]['status'])
        self.assertEqual("playing", status[1]['status'])
        self.assertEqual(self.mc.sounds["rainbow_disco_bears"].id, status[1]['sound_id'])

        self.mc.bcp_processor.send.reset_mock()
        self.advance_real_time(15)

        self.mc.bcp_processor.send.assert_has_calls([call('trigger', sound_instance=ANY,
                                                          name='playlist_playlist_sound_about_to_finish'),
                                                     call('trigger', name='attract_music_sound_changed'),
                                                     call('trigger', sound_instance=ANY,
                                                          name='dirty_grinding_beat_loop_played'),
                                                     call('trigger', sound_instance=ANY,
                                                          name='rainbow_disco_bears_stopped'),
                                                     call('trigger', sound_instance=ANY,
                                                          name='playlist_playlist_sound_stopped'),
                                                     call('trigger', name='attract_music_sound_stopped')])

        status = track_playlist.get_status()
        self.assertEqual("playing", status[0]['status'])
        self.assertEqual(self.mc.sounds["dirty_grinding_beat_loop"].id, status[0]['sound_id'])

        self.advance_real_time(2)
        self.mc.bcp_processor.send.reset_mock()

        self.mc.events.post('advance_playlist')
        self.advance_real_time(0.25)

        status = track_playlist.get_status()
        self.assertEqual("stopping", status[0]['status'])
        self.assertEqual(self.mc.sounds["dirty_grinding_beat_loop"].id, status[0]['sound_id'])
        self.assertEqual("playing", status[1]['status'])
        self.assertEqual(self.mc.sounds["hippie_ahead"].id, status[1]['sound_id'])

        self.advance_real_time(2)

        self.mc.bcp_processor.send.assert_has_calls([call('trigger', name='attract_music_sound_changed'),
                                                     call('trigger', sound_instance=ANY,
                                                          name='hippie_ahead_played'),
                                                     call('trigger', sound_instance=ANY,
                                                          name='dirty_grinding_beat_loop_stopped'),
                                                     call('trigger', sound_instance=ANY,
                                                          name='playlist_playlist_sound_stopped'),
                                                     call('trigger', name='attract_music_sound_stopped')])

        self.mc.bcp_processor.send.reset_mock()
        self.mc.events.post('stop_playlist')
        self.advance_real_time(3)

        self.mc.bcp_processor.send.assert_has_calls([call('trigger', sound_instance=ANY,
                                                          name='hippie_ahead_stopped'),
                                                     call('trigger', sound_instance=ANY,
                                                          name='playlist_playlist_sound_stopped'),
                                                     call('trigger', name='attract_music_sound_stopped'),
                                                     call('trigger', name='attract_music_stopped')])

        status = track_playlist.get_status()
        self.assertEqual("idle", status[0]['status'])
        self.assertEqual("idle", status[1]['status'])

    def test_playlist_context(self):
        """ Tests clearing Playlist context"""

        if SoundSystem is None or self.mc.sound_system is None:
            log = logging.getLogger('TestAudioPlaylist')
            log.warning("Sound system is not enabled - skipping audio tests")
            self.skipTest("Sound system is not enabled")

        self.assertIsNotNone(self.mc.sound_system)
        interface = self.mc.sound_system.audio_interface
        if interface is None:
            log = logging.getLogger('TestAudioPlaylist')
            log.warning("Sound system audio interface could not be loaded - skipping audio tests")
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(interface)

        # Check playlist track
        track_playlist = interface.get_track_by_name("playlist")
        self.assertIsNotNone(track_playlist)
        self.assertEqual("playlist", track_playlist.name)

        playlist_controller = interface.get_playlist_controller("playlist")
        self.assertIsNotNone(playlist_controller)
        self.assertEqual("playlist", playlist_controller.name)

        # Start the playlist playing with a specified context name
        playlist_instance = playlist_controller.play("attract_music",
                                                     "test_playlist_context",
                                                     {
                                                         "crossfade_time": 0.5,
                                                         "crossfade_mode": "override"
                                                     })
        self.assertIsNotNone(playlist_instance)
        self.assertEqual(0.5, playlist_instance.crossfade_time)
        self.advance_real_time()

        status = track_playlist.get_status()
        self.assertEqual("playing", status[0]['status'])
        self.assertEqual(self.mc.sounds["drumbeat_7"].id, status[0]['sound_id'])

        # Call clear context for an unknown context (playlist should keep playing)
        playlist_controller.clear_context("unknown_context")
        self.advance_real_time()
        status = track_playlist.get_status()
        self.assertEqual("playing", status[0]['status'])
        self.assertEqual(self.mc.sounds["drumbeat_7"].id, status[0]['sound_id'])

        # Call clear context using the context we started it playing (playlist should stop)
        playlist_controller.clear_context("test_playlist_context")
        self.advance_real_time()
        status = track_playlist.get_status()
        self.assertEqual("stopping", status[0]['status'])
        self.assertEqual(self.mc.sounds["drumbeat_7"].id, status[0]['sound_id'])

        # Allow time for playlist to fully stop
        self.advance_real_time(0.5)
        status = track_playlist.get_status()
        self.assertEqual("idle", status[0]['status'])
        self.assertEqual("idle", status[1]['status'])

        # Start the playlist playing again
        playlist_instance = playlist_controller.play("attract_music",
                                                     "ignore",
                                                     {
                                                         "crossfade_time": 0.5,
                                                         "crossfade_mode": "override"
                                                     })
        self.assertIsNotNone(playlist_instance)
        self.assertEqual(0.5, playlist_instance.crossfade_time)
        self.advance_real_time()

        status = track_playlist.get_status()
        self.assertEqual("playing", status[0]['status'])
        self.assertEqual(self.mc.sounds["drumbeat_7"].id, status[0]['sound_id'])

        # Start second playlist
        other_playlist_instance = playlist_controller.play("other_playlist",
                                                           "ignore",
                                                           {
                                                               "crossfade_time": 0.5,
                                                               "crossfade_mode": "override"
                                                           })
        self.assertIsNotNone(other_playlist_instance)
        self.assertEqual(0.5, other_playlist_instance.crossfade_time)
        self.advance_real_time()

        status = track_playlist.get_status()
        self.assertEqual("stopping", status[0]['status'])
        self.assertEqual(self.mc.sounds["drumbeat_7"].id, status[0]['sound_id'])
        self.assertEqual("playing", status[1]['status'])
        self.assertEqual(self.mc.sounds["hippie_ahead"].id, status[1]['sound_id'])

        # Start third playlist with a specific context (will be pending while other two playlists
        # are fading)
        third_playlist_instance = playlist_controller.play("third_playlist",
                                                           "test_pending_context",
                                                           {
                                                               "crossfade_time": 0.5,
                                                               "crossfade_mode": "override"
                                                           })
        self.assertIsNone(third_playlist_instance)
        self.assertTrue(playlist_controller.has_pending_request)
        self.advance_real_time()

        status = track_playlist.get_status()
        self.assertEqual("stopping", status[0]['status'])
        self.assertEqual(self.mc.sounds["drumbeat_7"].id, status[0]['sound_id'])
        self.assertEqual("playing", status[1]['status'])
        self.assertEqual(self.mc.sounds["hippie_ahead"].id, status[1]['sound_id'])

        # Call clear context using the context for the pending request (pending request should be cleared)
        playlist_controller.clear_context("test_pending_context")
        self.assertFalse(playlist_controller.has_pending_request)
        self.advance_real_time()

        status = track_playlist.get_status()
        self.assertEqual("stopping", status[0]['status'])
        self.assertEqual(self.mc.sounds["drumbeat_7"].id, status[0]['sound_id'])
        self.assertEqual("playing", status[1]['status'])
        self.assertEqual(self.mc.sounds["hippie_ahead"].id, status[1]['sound_id'])



"""Test custom code in MC."""
from mpfmc.tests.MpfIntegrationTestCase import MpfIntegrationTestCase


class TestCustomCode(MpfIntegrationTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'integration/machine_files/custom_code/'

    def test_asd(self):
        self.mock_event("my_return_event")

        # post event in MPF
        self.post_event("test_event")
        self.advance_time_and_run(.1)

        self.assertEventCalled("my_return_event")



from kivy.properties import NumericProperty
from kivy.uix.effectwidget import EffectBase


class GlowEffect(EffectBase):

    """GLSL effect to apply a glowing effect to a texture."""

    blur_size = NumericProperty(4.0)
    intensity = NumericProperty(.5)
    glow_amplitude = NumericProperty(0.1)
    glow_speed = NumericProperty(1.0)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.do_glsl()

    def on_blur_size(self, *args):
        self.do_glsl()

    def on_intensity(self, *args):
        self.do_glsl()

    def on_glow_speed(self, *args):
        self.do_glsl()

    def on_glow_amplitude(self, *args):
        self.do_glsl()

    def do_glsl(self):
        self.glsl = glow_glsl.format(float(self.blur_size),
                                     float(self.intensity),
                                     float(self.glow_amplitude),
                                     float(self.glow_speed))


glow_glsl = '''
vec4 effect(vec4 color, sampler2D texture, vec2 tex_coords, vec2 coords)
{{
    float blurSize = {0}/resolution.x;
    vec4 sum = vec4(0.0);
    sum += texture(texture, vec2(tex_coords.x - 4.0 * blurSize, tex_coords.y)) * .05;
    sum += texture(texture, vec2(tex_coords.x - 3.0*blurSize, tex_coords.y)) * 0.09;
    sum += texture(texture, vec2(tex_coords.x - 2.0*blurSize, tex_coords.y)) * 0.12;
    sum += texture(texture, vec2(tex_coords.x - blurSize, tex_coords.y)) * 0.15;
    sum += texture(texture, vec2(tex_coords.x, tex_coords.y)) * 0.16;
    sum += texture(texture, vec2(tex_coords.x + blurSize, tex_coords.y)) * 0.15;
    sum += texture(texture, vec2(tex_coords.x + 2.0*blurSize, tex_coords.y)) * 0.12;
    sum += texture(texture, vec2(tex_coords.x + 3.0*blurSize, tex_coords.y)) * 0.09;
    sum += texture(texture, vec2(tex_coords.x + 4.0*blurSize, tex_coords.y)) * 0.05;

    sum += texture(texture, vec2(tex_coords.x, tex_coords.y - 4.0*blurSize)) * 0.05;
    sum += texture(texture, vec2(tex_coords.x, tex_coords.y - 3.0*blurSize)) * 0.09;
    sum += texture(texture, vec2(tex_coords.x, tex_coords.y - 2.0*blurSize)) * 0.12;
    sum += texture(texture, vec2(tex_coords.x, tex_coords.y - blurSize)) * 0.15;
    sum += texture(texture, vec2(tex_coords.x, tex_coords.y)) * 0.16;
    sum += texture(texture, vec2(tex_coords.x, tex_coords.y + blurSize)) * 0.15;
    sum += texture(texture, vec2(tex_coords.x, tex_coords.y + 2.0*blurSize)) * 0.12;
    sum += texture(texture, vec2(tex_coords.x, tex_coords.y + 3.0*blurSize)) * 0.09;
    sum += texture(texture, vec2(tex_coords.x, tex_coords.y + 4.0*blurSize)) * 0.05;

    vec4 result = texture(texture, tex_coords);
    result = sum * ({2}*sin(2*3.14*{3}*time) + {1})  + result;
    return result;
}}
'''

effect_cls = GlowEffect
name = 'glow'



from mpf.config_players.plugin_player import PluginPlayer


class MpfWidgetPlayer(PluginPlayer):

    """Widget Player in MPF.

    Note: This class is loaded by MPF and everything in it is in the context of MPF.
    """

    config_file_section = 'widget_player'
    show_section = 'widgets'


def register_with_mpf(machine):
    """Register widget player in MPF module."""
    return 'widget', MpfWidgetPlayer(machine)


player_cls = MpfWidgetPlayer



"""Mission Pinball Framework Media Controller (mpf-mc) setup.py.

Notes:
    This setup script is a modified/customized version of the Kivy setup.py script.
"""

import sys
import re

from copy import deepcopy
import os
from os.path import join, dirname, sep, exists, isdir
from os import walk, environ
from distutils.version import LooseVersion
from distutils.sysconfig import get_python_inc
from collections import OrderedDict
from time import sleep

from sysconfig import get_paths
from setuptools import setup, Extension
print('Using setuptools')


# fix error with py3's LooseVersion comparisons
def ver_equal(self, other):
    return self.version == other


LooseVersion.__eq__ = ver_equal


MIN_CYTHON_STRING = '0.24'
MIN_CYTHON_VERSION = LooseVersion(MIN_CYTHON_STRING)
MAX_CYTHON_STRING = '0.29.21'
MAX_CYTHON_VERSION = LooseVersion(MAX_CYTHON_STRING)
CYTHON_UNSUPPORTED = (
    # ref https://github.com/cython/cython/issues/1968
    '0.27', '0.27.2'
)
CYTHON_REQUIRES_STRING = (
    'cython>={min_version},<={max_version},{exclusion}'.format(
        min_version=MIN_CYTHON_STRING,
        max_version=MAX_CYTHON_STRING,
        exclusion=','.join('!=%s' % excl for excl in CYTHON_UNSUPPORTED),
    )
)

PACKAGE_FILES_ALLOWED_EXT = ('py', 'yaml', 'png', 'md', 'zip', 'gif', 'jpg',
                             'mp4', 'm4v', 'so', 'pyd', 'dylib', 'wav', 'ogg',
                             'pxd', 'pyx', 'c', 'h', 'ttf', 'fnt', 'txt')

on_rtd = os.environ.get('READTHEDOCS') == 'True'


def getoutput(cmd, env=None):
    import subprocess
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE, env=env)
    p.wait()
    if p.returncode:  # if not returncode == 0
        print('WARNING: A problem occurred while running {0} (code {1})\n'
              .format(cmd, p.returncode))
        stderr_content = p.stderr.read()
        if stderr_content:
            print('{0}\n'.format(stderr_content))
        return ""
    return p.stdout.read()


def pkgconfig(*packages, **kw):
    flag_map = {'-I': 'include_dirs', '-L': 'library_dirs', '-l': 'libraries'}
    lenviron = None
    pconfig = join(sys.prefix, 'libs', 'pkgconfig')

    if isdir(pconfig):
        lenviron = environ.copy()
        lenviron['PKG_CONFIG_PATH'] = '{};{}'.format(
            environ.get('PKG_CONFIG_PATH', ''), pconfig)
    cmd = 'pkg-config --libs --cflags {}'.format(' '.join(packages))
    results = getoutput(cmd, lenviron).split()
    for token in results:
        extension = token[:2].decode('utf-8')
        flag = flag_map.get(extension)
        if not flag:
            continue
        kw.setdefault(flag, []).append(token[2:].decode('utf-8'))
    return kw


def get_isolated_env_paths():
    try:
        # sdl2_dev is installed before setup.py is run, when installing from
        # source due to pyproject.toml. However, it is installed to a
        # pip isolated env, which we need to add to compiler
        import kivy_deps.sdl2_dev as sdl2_dev
    except ImportError:
        return [], []

    sdl_root = os.path.abspath(join(sdl2_dev.__path__[0], '../../../..'))
    includes = [join(sdl_root, 'Include')] if isdir(join(sdl_root, 'Include')) else []
    libs = [join(sdl_root, 'libs')] if isdir(join(sdl_root, 'libs')) else []
    return includes, libs


# -----------------------------------------------------------------------------
# Determine on which platform we are

platform = sys.platform

# Detect 32/64bit for OSX (http://stackoverflow.com/a/1405971/798575)
if sys.platform == 'darwin':
    if sys.maxsize > 2 ** 32:
        osx_arch = 'x86_64'
    else:
        osx_arch = 'i386'

# Detect Python for android project (http://github.com/kivy/python-for-android)
ndkplatform = environ.get('NDKPLATFORM')
if ndkplatform is not None and environ.get('LIBLINK'):
    platform = 'android'
kivy_ios_root = environ.get('KIVYIOSROOT', None)
if kivy_ios_root is not None:
    platform = 'ios'
if exists('/opt/vc/include/bcm_host.h'):
    platform = 'rpi'
if exists('/usr/lib/arm-linux-gnueabihf/libMali.so'):
    platform = 'mali'

# -----------------------------------------------------------------------------
# Detect options
#
c_options = OrderedDict()
c_options['use_rpi'] = platform == 'rpi'
c_options['use_mali'] = platform == 'mali'
c_options['use_sdl2'] = True
c_options['use_gstreamer'] = True
c_options['use_avfoundation'] = platform == 'darwin'
c_options['use_osx_frameworks'] = platform == 'darwin'

# now check if environ is changing the default values
for key in list(c_options.keys()):
    ukey = key.upper()
    if ukey in environ:
        value = bool(int(environ[ukey]))
        print('Environ change {0} -> {1}'.format(key, value))
        c_options[key] = value


# -----------------------------------------------------------------------------
# Cython check
# Cython usage is optional (.c files are included to build without Cython)
#
cython_unsupported_append = '''

  Please note that the following versions of Cython are not supported
  at all: {}
'''.format(', '.join(map(str, CYTHON_UNSUPPORTED)))

cython_min = '''\
  This version of Cython is not compatible with MPF-MC. Please upgrade to
  at least version {0}, preferably the newest supported version {1}.

  If your platform provides a Cython package, make sure you have upgraded
  to the newest version. If the newest version available is still too low,
  please remove it and install the newest supported Cython via pip:

    pip install -I Cython=={1}{2}\
'''.format(MIN_CYTHON_STRING, MAX_CYTHON_STRING,
           cython_unsupported_append if CYTHON_UNSUPPORTED else '')

cython_max = '''\
  This version of Cython is untested with MPF-MC. While this version may
  work perfectly fine, it is possible that you may experience issues. If
  you do have issues, please downgrade to a supported version. It is
  best to use the newest supported version, {1}, but the minimum
  supported version is {0}.

  If your platform provides a Cython package, check if you can downgrade
  to a supported version. Otherwise, uninstall the platform package and
  install Cython via pip:

    pip install -I Cython=={1}{2}\
'''.format(MIN_CYTHON_STRING, MAX_CYTHON_STRING,
           cython_unsupported_append if CYTHON_UNSUPPORTED else '')

cython_unsupported = '''\
  This version of Cython suffers from known bugs and is unsupported.
  Please install the newest supported version, {1}, if possible, but
  the minimum supported version is {0}.

  If your platform provides a Cython package, check if you can install
  a supported version. Otherwise, uninstall the platform package and
  install Cython via pip:

    pip install -I Cython=={1}{2}\
'''.format(MIN_CYTHON_STRING, MAX_CYTHON_STRING,
           cython_unsupported_append)

have_cython = False
skip_cython = environ.get('USE_CYTHON', False) not in ['1', 'True', 'TRUE', 'true', 'Yes', 'YES', 'y', 'Y']

if skip_cython:
    print("\nSkipping Cython build (using .c files)")
else:
    try:
        # check for cython
        from Cython.Distutils import build_ext
        have_cython = True
        import Cython
        cy_version_str = Cython.__version__
        cy_ver = LooseVersion(cy_version_str)
        print('\nDetected Cython version {}'.format(cy_version_str))
        if cy_ver < MIN_CYTHON_VERSION:
            print(cython_min)
            raise ImportError('Incompatible Cython Version')
        if cy_ver in CYTHON_UNSUPPORTED:
            print(cython_unsupported)
            raise ImportError('Incompatible Cython Version')
        if cy_ver > MAX_CYTHON_VERSION:
            print(cython_max)
            sleep(1)
    except ImportError:
        print("\nCython is missing and the USE_CYTHON environment variable is set to True!\n\n")
        raise

if not have_cython:
    from distutils.command.build_ext import build_ext   # noqa

# -----------------------------------------------------------------------------
# Setup classes

# the build path where kivy is being compiled
src_path = build_path = dirname(__file__)


class CustomBuildExt(build_ext):

    def finalize_options(self):
        retval = build_ext.finalize_options(self)
        global build_path   # noqa
        if (self.build_lib is not None and exists(self.build_lib) and
                not self.inplace):
            build_path = self.build_lib
        return retval

    def build_extensions(self):
        c = self.compiler.compiler_type
        print('Detected compiler is {}'.format(c))
        if c != 'msvc':
            for e in self.extensions:
                e.extra_link_args += ['-lm']

        build_ext.build_extensions(self)


def _check_and_fix_sdl2_mixer(f_path_to_check):
    # Between SDL_mixer 2.0.1 and 2.0.4, the included frameworks changed
    # smpeg2 have been replaced with mpg123, but there is no need to fix.
    smpeg2_path = ("{}/Versions/A/Frameworks/smpeg2.framework"
                   "/Versions/A/smpeg2").format(f_path_to_check)
    if not exists(smpeg2_path):
        return

    print("Check if SDL2_mixer smpeg2 have an @executable_path")
    rpath_from = ("@executable_path/../Frameworks/SDL2.framework"
                  "/Versions/A/SDL2")
    rpath_to = "@rpath/../../../../SDL2.framework/Versions/A/SDL2"
    output = getoutput(("otool -L '{}'").format(smpeg2_path)).decode('utf-8')
    if "@executable_path" not in output:
        return

    print("WARNING: Your SDL2_mixer version is invalid")
    print("WARNING: The smpeg2 framework embedded in SDL2_mixer contains a")
    print("WARNING: reference to @executable_path that will fail the")
    print("WARNING: execution of your application.")
    print("WARNING: We are going to change:")
    print("WARNING: from: {}".format(rpath_from))
    print("WARNING: to: {}".format(rpath_to))
    getoutput("install_name_tool -change {} {} {}".format(
        rpath_from, rpath_to, smpeg2_path))

    output = getoutput(("otool -L '{}'").format(smpeg2_path))
    if b"@executable_path" not in output:
        print("WARNING: Change successfully applied!")
        print("WARNING: You'll never see this message again.")
    else:
        print("WARNING: Unable to apply the changes, sorry.")


gst_flags = {}

if platform == 'darwin':
    if c_options['use_osx_frameworks']:
        if osx_arch == "i386":
            print("Warning: building with frameworks fail on i386")
        else:
            print("OSX framework used, force to x86_64 only")
            environ["ARCHFLAGS"] = environ.get("ARCHFLAGS", "-arch x86_64")
            print("OSX ARCHFLAGS are: {}".format(environ["ARCHFLAGS"]))

# detect gstreamer, only on desktop
# works if we forced the options or in autodetection
if c_options['use_gstreamer'] in (None, True):
    gstreamer_valid = False
    if c_options['use_osx_frameworks'] and platform == 'darwin':
        # check the existence of frameworks
        f_path = '/Library/Frameworks/GStreamer.framework'
        if not exists(f_path):
            c_options['use_gstreamer'] = False
            print('GStreamer framework not found, fallback on pkg-config')
        else:
            print('GStreamer framework found')
            gstreamer_valid = True
            c_options['use_gstreamer'] = True
            gst_flags = {
                'extra_link_args': [
                    '-F/Library/Frameworks',
                    '-Xlinker', '-rpath',
                    '-Xlinker', '/Library/Frameworks',
                    '-Xlinker', '-headerpad',
                    '-Xlinker', '190',
                    '-framework', 'GStreamer'],
                'include_dirs': [join(f_path, 'Headers')]}
    elif platform == 'win32':
        gst_flags = pkgconfig('gstreamer-1.0')
        if 'libraries' in gst_flags:
            print('GStreamer found via pkg-config')
            gstreamer_valid = True
            c_options['use_gstreamer'] = True
        else:
            _includes = get_isolated_env_paths()[0] + [get_paths()['include']]
            for include_dir in _includes:
                if exists(join(include_dir, 'gst', 'gst.h')):
                    print('GStreamer found via gst.h')
                    gstreamer_valid = True
                    c_options['use_gstreamer'] = True
                    gst_flags = {
                        'libraries':
                            ['gstreamer-1.0', 'glib-2.0', 'gobject-2.0']}
                    break

    if not gstreamer_valid:
        # use pkg-config approach instead
        gst_flags = pkgconfig('gstreamer-1.0')
        if 'libraries' in gst_flags:
            print('GStreamer found via pkg-config')
            gstreamer_valid = True
            c_options['use_gstreamer'] = True

    if not gstreamer_valid:
        raise RuntimeError('GStreamer not found and is required to build MPF-MC')

# detect SDL2
# works if we forced the options or in autodetection
sdl2_flags = {}
if c_options['use_sdl2'] in (None, True):
    sdl2_valid = False
    if c_options['use_osx_frameworks'] and platform == 'darwin':
        # check the existence of frameworks
        sdl2_valid = True
        sdl2_flags = {
            'extra_link_args': [
                '-F/Library/Frameworks',
                '-Xlinker', '-rpath',
                '-Xlinker', '/Library/Frameworks',
                '-Xlinker', '-headerpad',
                '-Xlinker', '190'],
            'include_dirs': [],
            'extra_compile_args': ['-F/Library/Frameworks']
        }
        for name in ('SDL2', 'SDL2_image', 'SDL2_mixer'):
            f_path = '/Library/Frameworks/{}.framework'.format(name)
            if not exists(f_path):
                print('Missing framework {}'.format(f_path))
                sdl2_valid = False
                continue
            sdl2_flags['extra_link_args'] += ['-framework', name]
            sdl2_flags['include_dirs'] += [join(f_path, 'Headers')]
            print('Found sdl2 frameworks: {}'.format(f_path))
            if name == 'SDL2_mixer':
                _check_and_fix_sdl2_mixer(f_path)

        if not sdl2_valid:
            c_options['use_sdl2'] = False
            print('SDL2 frameworks not found, fallback on pkg-config')
        else:
            c_options['use_sdl2'] = True
            print('Activate SDL2 compilation')

    if not sdl2_valid and platform != "ios":
        # use pkg-config approach instead
        sdl2_flags = pkgconfig('sdl2', 'SDL2_image', 'SDL2_mixer')
        if 'libraries' in sdl2_flags:
            print('SDL2 found via pkg-config')
            c_options['use_sdl2'] = True

# -----------------------------------------------------------------------------
# declare flags


def get_modulename_from_file(filename_to_check):
    filename_to_check = filename_to_check.replace(sep, '/')
    pyx = '.'.join(filename_to_check.split('.')[:-1])
    pyxl = pyx.split('/')
    while pyxl[0] != 'mpfmc':
        pyxl.pop(0)
    if pyxl[1] == 'mpfmc':
        pyxl.pop(0)
    return '.'.join(pyxl)


def expand(root_path, *args):
    return join(root_path, 'mpfmc', *args)


class CythonExtension(Extension):

    def __init__(self, *args, **kwargs):
        Extension.__init__(self, *args, **kwargs)
        self.cython_directives = {
            'c_string_encoding': 'utf-8',
            'profile': 'USE_PROFILE' in environ,
            'embedsignature': environ.get('USE_EMBEDSIGNATURE', '0') == 1,
            'language_level': 3,
            'unraisable_tracebacks': True}
        # XXX with pip, setuptools is imported before distutils, and change
        # our pyx to c, then, cythonize doesn't happen. So force again our
        # sources
        self.sources = args[1]


def merge(d1, *args):
    d1 = deepcopy(d1)
    for d2 in args:
        for item_key, item_value in d2.items():
            item_value = deepcopy(item_value)
            if item_key in d1:
                d1[item_key].extend(item_value)
            else:
                d1[item_key] = item_value
    return d1


def determine_base_flags():
    flags = {
        'libraries': [],
        'include_dirs': [join(src_path, 'kivy', 'include')],
        'library_dirs': [],
        'extra_link_args': [],
        'extra_compile_args': []}
    if platform.startswith('freebsd'):
        flags['include_dirs'] += [join(
            environ.get('LOCALBASE', '/usr/local'), 'include')]
        flags['library_dirs'] += [join(
            environ.get('LOCALBASE', '/usr/local'), 'lib')]
    elif platform == 'darwin':
        v = os.uname()
        if v[2] >= '13.0.0':
            # use xcode-select to search on the right Xcode path
            # XXX use the best SDK available instead of a specific one
            import platform as _platform
            xcode_dev = getoutput('xcode-select -p').splitlines()[0]
            sdk_mac_ver = '.'.join(_platform.mac_ver()[0].split('.')[:2])
            print('Xcode detected at {}, and using OS X{} sdk'.format(
                xcode_dev, sdk_mac_ver))
            sysroot = join(
                xcode_dev.decode('utf-8'),
                'Platforms/MacOSX.platform/Developer/SDKs',
                'MacOSX{}.sdk'.format(sdk_mac_ver),
                'System/Library/Frameworks')
        else:
            sysroot = ('/System/Library/Frameworks/'
                       'ApplicationServices.framework/Frameworks')
        flags['extra_compile_args'] += ['-F%s' % sysroot]
        flags['extra_link_args'] += ['-F%s' % sysroot]
    elif platform == 'win32':
        flags['include_dirs'] += [get_python_inc(prefix=sys.prefix)]
        flags['library_dirs'] += [join(sys.prefix, "libs")]
    return flags


def determine_sdl2():
    flags = {}
    if not c_options['use_sdl2']:
        return flags

    sdl2_path = environ.get('KIVY_SDL2_PATH', None)

    if sdl2_flags and not sdl2_path and platform == 'darwin':
        return sdl2_flags

    includes, _ = get_isolated_env_paths()

    # no pkgconfig info, or we want to use a specific sdl2 path, so perform
    # manual configuration
    flags['libraries'] = ['SDL2', 'SDL2_image', 'SDL2_mixer']
    split_chr = ';' if platform == 'win32' else ':'
    sdl2_paths = sdl2_path.split(split_chr) if sdl2_path else []

    if not sdl2_paths:
        sdl2_paths = []
        for include in includes + [join(sys.prefix, 'include')]:
            sdl_inc = join(include, 'SDL2')
            if isdir(sdl_inc):
                sdl2_paths.append(sdl_inc)
        sdl2_paths.extend(['/usr/local/include/SDL2', '/usr/include/SDL2'])

    flags['include_dirs'] = sdl2_paths
    flags['extra_link_args'] = []
    flags['extra_compile_args'] = []
    flags['library_dirs'] = (
        sdl2_paths if sdl2_paths else
        ['/usr/local/lib/'])

    if sdl2_flags:
        flags = merge(flags, sdl2_flags)

    # ensure headers for all the SDL2 and sub libraries are available
    libs_to_check = ['SDL', 'SDL_mixer', 'SDL_image']
    can_compile = True
    for lib in libs_to_check:
        found = False
        for d in flags['include_dirs']:
            inc_dir = join(d, '{}.h'.format(lib))
            if exists(inc_dir):
                found = True
                print('SDL2: found {} header at {}'.format(lib, inc_dir))
                break

        if not found:
            print('SDL2: missing sub library {}'.format(lib))
            can_compile = False

    if not can_compile:
        c_options['use_sdl2'] = False
        return {}

    return flags


base_flags = determine_base_flags()

# -----------------------------------------------------------------------------
# sources to compile
sources = {
    'core/audio/sound_file.pyx': {
        'depends': ['core/audio/sdl2_helper.h', 'core/audio/gstreamer_helper.h']},
    'core/audio/track.pyx': {
        'depends': ['core/audio/sdl2_helper.h', 'core/audio/gstreamer_helper.h']},
    'core/audio/track_standard.pyx': {
        'depends': ['core/audio/sdl2_helper.h', 'core/audio/gstreamer_helper.h']},
    'core/audio/track_sound_loop.pyx': {
        'depends': ['core/audio/sdl2_helper.h', 'core/audio/gstreamer_helper.h']},
    'core/audio/audio_interface.pyx': {
        'depends': ['core/audio/sdl2_helper.h', 'core/audio/gstreamer_helper.h']},
    'core/audio/playlist_controller.pyx': {},
    'uix/bitmap_font/bitmap_font.pyx': {'depends': ['core/audio/sdl2.pxi', ]}
}

if c_options["use_sdl2"] and not on_rtd:
    sdl2_flags = determine_sdl2()
else:
    sdl2_flags = {}

if sdl2_flags:
    for source_file, depends in sources.items():
        sources[source_file] = merge(
            base_flags, gst_flags, sdl2_flags, depends)


# -----------------------------------------------------------------------------
# extension modules

def get_extensions_from_sources(sources_to_search):
    ext_modules_found = []
    for pyx, flags in sources_to_search.items():
        pyx = expand(src_path, pyx)
        depends_sources = [expand(src_path, x) for x in flags.pop('depends', [])]
        c_depends = [expand(src_path, x) for x in flags.pop('c_depends', [])]
        if not have_cython:
            pyx = '%s.c' % pyx[:-4]
        f_depends = [x for x in depends_sources if x.rsplit('.', 1)[-1] in (
            'c', 'cpp', 'm')]
        module_name = get_modulename_from_file(pyx)
        flags_clean = {'depends': depends_sources}
        for item_key, item_value in flags.items():
            if item_value:
                flags_clean[item_key] = item_value
        ext_modules_found.append(CythonExtension(
            module_name, [pyx] + f_depends + c_depends, **flags_clean))
    return ext_modules_found


print(sources)

if not on_rtd:
    ext_modules = get_extensions_from_sources(sources)
else:
    ext_modules = []

# -----------------------------------------------------------------------------
# Get the version number of mpf-mc and the required version of MPF by reading
# the file directly. We can't import it because that would import mpf and
# break the setup. Details here:
# http://stackoverflow.com/questions/458550/standard-way-to-embed-version
# -into-python-package
version_file = "mpfmc/_version.py"
version_file_content = open(version_file, "rt").read()
version_re = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(version_re, version_file_content, re.M)
if mo:
    mc_version = mo.group(1)
else:
    raise RuntimeError(
        "Unable to find version string in %s." % (version_file,))

# This section pulls the MPF required version from the mpf-mc version file so
# we can write that as a requirement below
mpf_version_re = r"^__mpf_version_required__ = ['\"]([^'\"]*)['\"]"
mo = re.search(mpf_version_re, version_file_content, re.M)
if mo:
    mpf_version = mo.group(1)
else:
    raise RuntimeError("Unable to find MPF version string in %s." % (
        version_file,))

install_requires = ['ruamel.yaml==0.15.100',  # better YAML library
                    'mpf>={}'.format(mpf_version),
                    'kivy==1.11.1',
                    'psutil==5.7.0',
                    'Pygments==2.3.1',  # YAML syntax formatting for the iMC
                    'pypiwin32==223;platform_system=="Windows"',
                    # also update those in appveyor.yaml if you change versions
                    'kivy_deps.sdl2==0.1.23;platform_system=="Windows"',
                    'kivy_deps.sdl2-dev==0.1.23;platform_system=="Windows"',
                    'kivy_deps.glew==0.1.12;platform_system=="Windows"',
                    'kivy_deps.glew-dev==0.1.12;platform_system=="Windows"',
                    'kivy_deps.gstreamer==0.1.18;platform_system=="Windows"',
                    'kivy_deps.gstreamer-dev==0.1.18;platform_system=="Windows"',
                    'ffpyplayer==4.3.1'
                    ]

# If we're running on Read The Docs, then we just need to copy the files
# (since mpf-docs uses the test YAML files in the doc build), and we don't
# need to actually install mpf-mc, so override the installation requirements:

if on_rtd:
    install_requires = []

# -----------------------------------------------------------------------------
# automatically detect package files
package_files = dict(mpfmc=list())
for root, subFolders, files in walk('mpfmc'):
    for fn in files:
        ext = fn.split('.')[-1].lower()
        if ext not in PACKAGE_FILES_ALLOWED_EXT:
            continue

        filename = join(root, fn)
        directory = dirname(filename)
        package_files['mpfmc'].append('/'.join(filename.split(os.sep)[1:]))


# -----------------------------------------------------------------------------
# setup !
setup(
    name='mpf-mc',
    version=mc_version,
    author='The Mission Pinball Framework Team',
    author_email='brian@missionpinball.org',
    url='http://missionpinball.org',
    license='MIT',
    description='Mission Pinball Framework Media Controller',
    long_description='''Graphics, video, and audio engine for the
        Mission Pinball Framework.

        The Mission Pinball Framework Media Controller (MPF-MC) is a component
        of the Mission Pinball Framework (MPF) that controls graphics and
        sound, including dot matrix displays (DMDs), LCD displays, and color
        RGB LED displays.

        (The MPF media controller architecture is modular, so you can use this
        MPF-MC package or another one.)

        The MPF-MC is built on Kivy and leverages SDL2, OpenGL, and
        GPU-accelerated hardware.

        MPF is a work-in-progress that is not yet complete, though we're
        actively developing it and checking in several commits a week. It's
        MIT licensed, actively developed by fun people, and supported by a
        vibrant pinball-loving community.''',
    keywords='pinball',
    ext_modules=ext_modules,
    cmdclass={'build_ext': CustomBuildExt},
    packages=['mpfmc'],
    package_dir={'mpfmc': 'mpfmc'},
    package_data=package_files,
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Topic :: Artistic Software',
        'Topic :: Games/Entertainment :: Arcade'
    ],
    install_requires=install_requires,
    tests_require=[],
    entry_points='''
    [mpf.config_player]
    sound_player=mpfmc.config_players.plugins.sound_player:register_with_mpf
    sound_loop_player=mpfmc.config_players.plugins.sound_loop_player:register_with_mpf
    playlist_player=mpfmc.config_players.plugins.playlist_player:register_with_mpf
    widget_player=mpfmc.config_players.plugins.widget_player:register_with_mpf
    slide_player=mpfmc.config_players.plugins.slide_player:register_with_mpf
    track_player=mpfmc.config_players.plugins.track_player:register_with_mpf
    display_light_player=mpfmc.config_players.plugins.display_light_player:register_with_mpf

    [mpf.command]
    mc=mpfmc.commands.mc:get_command
    imc=mpfmc.commands.imc:get_command
    ''',
    setup_requires=[CYTHON_REQUIRES_STRING] if not skip_cython else [])



from kivy.properties import OptionProperty

from mpfmc.uix.transitions import MpfTransition


class MoveInTransition(MpfTransition):
    """Move In Transition, the current slide does not move, and the
    incoming slide "moves in" on top it."""

    direction = OptionProperty('left', options=('left', 'right', 'top',
                                                'bottom'))

    def on_progress(self, progression):

        s_in, s_out, width, height, progress = self.get_vars(progression)

        del s_out

        direction = self.direction

        if direction == 'left':
            s_in.x = width * (1 - progress)

        elif direction == 'right':
            s_in.x = 0 - width * (1 - progress)

        elif direction == 'top':
            s_in.y = height * (1 - progress)

        elif direction == 'bottom':
            s_in.y = 0 - height * (1 - progress)


transition_cls = MoveInTransition
name = 'move_in'



from mpfmc.core.audio.sdl2 cimport *
from mpfmc.core.audio.gstreamer cimport *


ctypedef struct SampleMemory:
    gpointer data
    gsize size

ctypedef struct SampleStream:
    GstElement *pipeline
    GstElement *sink
    GstSample *sample
    GstBuffer *buffer
    GstMapInfo map_info
    Uint32 map_buffer_pos
    gboolean map_contains_valid_sample_data
    gint null_buffer_count

cdef enum SoundType:
    sound_type_memory = 0
    sound_type_streaming = 1

ctypedef union SoundSampleData:
    SampleMemory *memory
    SampleStream *stream

ctypedef struct SoundSample:
    SoundType type
    SoundSampleData data
    double duration
    Uint32 size


cdef class SoundFile:
    """SoundFile is the base class for wrapper classes used to manage sound sample data."""
    cdef str file_name
    cdef AudioCallbackData *callback_data
    cdef SoundSample sample
    cdef object log


cdef class SoundMemoryFile(SoundFile):
    """SoundMemoryFile is a wrapper class to manage sound sample data stored
    in memory."""
    cdef bint _loaded_using_sdl


cdef class SoundStreamingFile(SoundFile):
    """SoundStreamingFile is a wrapper class to manage streaming sound sample data."""
    cdef GstElement *pipeline
    cdef GstElement *source
    cdef GstElement *convert
    cdef GstElement *resample
    cdef GstElement *sink
    cdef GstBus *bus
    cdef gulong bus_message_handler_id



"""Text Input widget.

The :class:`MpfTextInput` widget is used to all the player (or an operator) to
enter text. It can be linked to a :class:`Text` widget which displays the text
that's been entered so far.

"""
from collections import deque
from typing import Optional
from kivy.clock import Clock

from mpfmc.uix.widget import WidgetContainer
from mpfmc.widgets.text import Text

MYPY = False
if MYPY:   # pragma: no cover
    from mpfmc.core.mc import MpfMc


class MpfTextInput(Text):

    """Text input widget."""

    widget_type_name = 'text_input'
    animation_properties = list()

    def __init__(self, mc: "MpfMc", config: dict, key: Optional[str] = None, **kwargs) -> None:
        """Initialise text input.

        Note that this class is called *MpfTextInput* instead of *TextInput*
        because Kivy has a class called *TextInput* which collides with this
        one if they have the same name.
        """
        super().__init__(mc=mc, config=config, key=key)

        del kwargs

        self.linked_text_widget = None
        self.linked_text_widget_right_edge = None
        self.active = False
        self.registered_event_handlers = set()

        # setup the char list and set the current displayed char
        self.char_list = deque()
        self.char_list.extend(self.config['char_list'])
        self.char_list.append('back')
        self.char_list.append('end')
        self.current_list = self.char_list
        self.final_list = deque()
        self.final_list.append('back')
        self.final_list.append('end')
        self._is_blocking = False

        Clock.schedule_once(self.find_linked_text_widget)

    def __repr__(self) -> str:
        try:
            return '<TextInput Widget key={}>'.format(self.key)
        except AttributeError:
            return '<TextInput Widget>'

    def find_linked_text_widget(self, dt) -> None:
        del dt

        for target in self.mc.targets.values():
            w = target.find_widgets_by_key(self.config['key'])
            if w:
                if w[0] != self and isinstance(w[0], Text):
                    self.linked_text_widget = w[0]
                elif len(w) > 1 and isinstance(w[1], Text):
                    self.linked_text_widget = w[1]
                break

        if self.linked_text_widget:
            self.active = True
            self._register_events()

            if self.config['dynamic_x']:
                # bind() is weak meth, so we don't have to unbind
                self.linked_text_widget.bind(size=self.set_relative_position)
                self.linked_text_widget_right_edge = (
                    self.linked_text_widget.get_text_width() + self.linked_text_widget.x)

        self.jump(self.config['initial_char'])

    def _register_events(self) -> None:
        if not self.config['shift_left_event']:
            self.config['shift_left_event'] = (
                'text_input_{}_shift_left'.format(self.key))

        if not self.config['shift_right_event']:
            self.config['shift_right_event'] = (
                'text_input_{}_shift_right'.format(self.key))

        if not self.config['select_event']:
            self.config['select_event'] = (
                'text_input_{}_select'.format(self.key))

        if not self.config['abort_event']:
            self.config['abort_event'] = (
                'text_input_{}_abort'.format(self.key))

        if not self.config['force_complete_event']:
            self.config['force_complete_event'] = (
                'text_input_{}_force_complete'.format(self.key))

        for event in self.config.get('block_events', []):
            self.registered_event_handlers.add(self.mc.events.add_handler(
                event, self.block_enable, priority=2  # Priority avoids conflicts with select/abort event handlers
            ))
            self.mc.bcp_processor.register_trigger(event)
        for event in self.config.get('release_events', []):
            self.registered_event_handlers.add(self.mc.events.add_handler(
                event, self.block_disable, priority=2
            ))
            self.mc.bcp_processor.register_trigger(event)

        self.registered_event_handlers.add(self.mc.events.add_handler(
            self.config['shift_left_event'], self.shift, places=-1))
        self.mc.bcp_processor.register_trigger(self.config['shift_left_event'])
        self.registered_event_handlers.add(self.mc.events.add_handler(
            self.config['shift_right_event'], self.shift, places=1))
        self.mc.bcp_processor.register_trigger(self.config['shift_right_event'])
        self.registered_event_handlers.add(self.mc.events.add_handler(
            self.config['select_event'], self.select))
        self.mc.bcp_processor.register_trigger(self.config['select_event'])
        self.registered_event_handlers.add(self.mc.events.add_handler(
            self.config['abort_event'], self.abort))
        self.mc.bcp_processor.register_trigger(self.config['abort_event'])
        self.registered_event_handlers.add(self.mc.events.add_handler(
            self.config['force_complete_event'], self.complete))
        self.mc.bcp_processor.register_trigger(self.config['force_complete_event'])

    def _deregister_events(self) -> None:
        self.mc.events.remove_handlers_by_keys(self.registered_event_handlers)

    def jump(self, char: str) -> None:
        # Unfortunately deque.index() is Python 3.5+, so we have to do it this
        # way.

        if not self.current_list:
            return

        counts_remaining = len(self.current_list)
        while self.current_list[0] != char:
            self.current_list.rotate(1)
            counts_remaining -= 1

            if not counts_remaining:
                raise ValueError("Cannot set text_input character to "
                                 "{} as that is not a valid entry in the "
                                 "widget's character list".format(char))

        self.shift(0, True)

    def shift(self, places: int = 1, force: bool = False, **kwargs) -> None:
        del kwargs
        if self._is_blocking:
            return
        if self.active or force:
            self.current_list.rotate(-places)

            if self.current_list[0] == 'end':
                self.font_size = self.config['font_size'] / 2
                self.update_text('END')
            elif self.current_list[0] == ' ':
                self.font_size = self.config['font_size'] / 2
                self.update_text('SPACE')
            elif self.current_list[0] == 'back':
                self.font_size = self.config['font_size'] / 2
                self.update_text('BACK')
            else:
                self.font_size = self.config['font_size']
                self.update_text(self.current_list[0])

            if self.config['dynamic_x'] and self.linked_text_widget:
                self.set_relative_position()

    def select(self, **kwargs) -> None:
        del kwargs
        if not self.active:
            return

        if self.current_list[0] == 'back':

            if self.linked_text_widget.text:
                # set the text_entry text to whatever the last char was
                self.current_list = self.char_list
                self.jump(self.linked_text_widget.text[-1:])
                # remove the last char from the associated widget
                self.linked_text_widget.update_text(
                    self.linked_text_widget.text[:-1])

        elif self.current_list[0] == 'end':
            self.complete()

        else:
            self.linked_text_widget.update_text(
                self.linked_text_widget.text + self.current_list[0])

            if len(self.linked_text_widget.text) >= self.config['max_chars']:
                # only show back and end
                self.current_list = self.final_list
                self.jump("end")

            if len(self.linked_text_widget.text) > self.config['max_chars']:
                # we are done
                self.complete()

    def set_relative_position(self, *args) -> None:
        del args

        new_right_edge = (self.linked_text_widget.get_text_width() +
                          self.linked_text_widget.x)

        self.x += new_right_edge - self.linked_text_widget_right_edge
        self.linked_text_widget_right_edge = new_right_edge

        if self.x < (self.linked_text_widget_right_edge +
                     self.config['dynamic_x_pad']):
            self.x += (self.linked_text_widget_right_edge +
                       self.config['dynamic_x_pad'] - self.x)

    def complete(self, **kwargs) -> None:
        del kwargs
        self.done()
        self.mc.post_mc_native_event('text_input_{}_complete'.format(self.key),
                                     text=self.linked_text_widget.text)
        """event: text_input_(key)_complete

        desc: This event is posted by a *text_input* display widget when the
            entered text is finalized.

        args:
            text: A string of the final characters that were entered.
        """

    def abort(self, **kwargs) -> None:
        del kwargs
        self.done()
        self.mc.post_mc_native_event('text_input_{}_abort'.format(self.key),
                                     text=self.linked_text_widget.text)
        """event: text_input_(key)_abort

        desc: This event is posted by a *text_input* display widget when the
            entering process was aborted.

        args:
            text: A string of the characters that were entered so far.
        """

    def done(self) -> None:
        self._deregister_events()
        self.active = False
        self.text = ''
        self._container.parent.remove_widget(self)

    def prepare_for_removal(self) -> None:
        self.done()
        super().prepare_for_removal()

    def block_enable(self, **kwargs) -> None:
        del kwargs
        self._is_blocking = True

    def block_disable(self, **kwargs) -> None:
        del kwargs
        self._is_blocking = False


widget_classes = [MpfTextInput]



"""BCP Server interface for the MPF Media Controller"""

import logging
import queue
import socket
import sys
import threading
import time
import traceback

import select

import mpf.core.bcp.bcp_socket_client as bcp
from mpf.exceptions.runtime_error import MpfRuntimeError


class BCPServer(threading.Thread):
    """Parent class for the BCP Server thread.

    Args:
        mc: A reference to the main MediaController instance.
        receiving_queue: A shared Queue() object which holds incoming BCP
            commands.
        sending_queue: A shared Queue() object which holds outgoing BCP
            commands.

    """

    def __init__(self, mc, receiving_queue, sending_queue):

        threading.Thread.__init__(self)
        self.mc = mc
        self.log = logging.getLogger('MPF-MC BCP Server')
        self.receive_queue = receiving_queue
        self.sending_queue = sending_queue
        self.connection = None
        self.socket = None
        self.done = False

        self.setup_server_socket(mc.machine_config['mpf-mc']['bcp_interface'],
                                 mc.machine_config['mpf-mc']['bcp_port'])
        self.sending_thread = threading.Thread(target=self.sending_loop)
        self.sending_thread.daemon = True
        self.sending_thread.start()

    def setup_server_socket(self, interface='localhost', port=5050):
        """Sets up the socket listener.

        Args:
            interface: String name of which interface this socket will listen
                on.
            port: Integer TCP port number the socket will listen on.

        """
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        self.log.info('Starting up on %s port %s', interface, port)

        try:
            self.socket.bind((interface, port))
        except IOError as e:
            raise MpfRuntimeError("Failed to bind BCP Socket to {} on port {}. "
                                  "Is there another application running on that port?".format(interface, port), 1,
                                  self.log.name) from e

        self.socket.listen(5)
        self.socket.settimeout(1)

    def run(self):
        """The socket thread's run loop."""
        try:
            while not self.mc.thread_stopper.is_set():
                self.log.info("Waiting for a connection...")
                # Since posting an event from a thread is not safe, we just
                # drop the event we want into the receive queue and let the
                # main loop pick it up
                self.receive_queue.put(('trigger',
                                        {'name': 'client_disconnected',
                                         'host': self.socket.getsockname()[0],
                                         'port': self.socket.getsockname()[1]}))
                '''event: client_disconnected
                desc: Posted on the MPF-MC only (e.g. not in MPF) when the BCP
                client disconnects. This event is also posted when the MPF-MC
                starts before a client is connected.

                This is useful for triggering a slide notifying of the
                disconnect.

                args:
                host: The hostname or IP address that the socket is listening
                on.
                port: The port that the socket is listening on.

                '''
                self.mc.bcp_client_connected = False

                start_time = time.time()
                while (not self.connection and
                        not self.mc.thread_stopper.is_set()):
                    try:
                        self.connection, client_address = self.socket.accept()
                    except (socket.timeout, OSError):
                        if self.mc.options['production'] and start_time + 30 < time.time():
                            self.log.warning("Timeout while waiting for connection. Stopping!")
                            self.mc.stop()
                            return
                        if self.mc.thread_stopper.is_set():
                            self.log.info("Stopping BCP listener thread")
                            return

                self.log.info("Received connection from: %s:%s",
                              client_address[0], client_address[1])

                # Since posting an event from a thread is not safe, we just
                # drop the event we want into the receive queue and let the
                # main loop pick it up
                self.receive_queue.put(('trigger',
                                        {'name': 'client_connected',
                                         'host': client_address[0],
                                         'port': client_address[1]}))

                '''event: client_connected
                desc: Posted on the MPF-MC only when a BCP client has
                connected.

                args:
                address: The IP address of the client that connected.
                port: The port the client connected on.
                '''

                self.mc.bcp_client_connected = True
                socket_chars = b''

                # Receive the data in small chunks and retransmit it
                while not self.mc.thread_stopper.is_set():
                    # todo Better large message handling
                    # if a json message comes in that's larger than 8192, it
                    # won't be processed properly. So if we end up sticking
                    # with BCP, we should probably change this to be smarter
                    # and to check for complete messages before processing them

                    ready = select.select([self.connection], [], [], 1)
                    if ready[0]:
                        try:
                            data_read = self.connection.recv(8192)
                        except socket.timeout:
                            pass

                        if data_read:
                            socket_chars += data_read
                            commands = socket_chars.split(b"\n")

                            # keep last incomplete command
                            socket_chars = commands.pop()

                            # process all complete commands
                            self._process_receives_messages(commands)
                        else:
                            # no bytes -> socket closed
                            break

                # close connection. while loop will not exit if this is not intended.
                self.connection.close()
                self.connection = None

                # always exit
                self.mc.stop()
                return

        except Exception:   # noqa
            exc_type, exc_value, exc_traceback = sys.exc_info()
            lines = traceback.format_exception(exc_type, exc_value,
                                               exc_traceback)
            msg = ''.join(line for line in lines)
            self.mc.crash_queue.put(msg)

    def _process_receives_messages(self, commands):
        # process all complete commands
        for cmd in commands:
            if cmd:
                try:
                    decoded_cmd = cmd.strip().decode()
                except UnicodeDecodeError:
                    self.log.warning("Failed to decode BCP message: %s", cmd.strip())
                    continue

                self.process_received_message(decoded_cmd)

    def stop(self):
        """ Stops and shuts down the BCP server."""
        if not self.done:
            self.log.info("Socket thread stopping.")
            self.sending_queue.put('goodbye', None)
            time.sleep(1)  # give it a chance to send goodbye before quitting
            self.done = True
            self.mc.done = True

    def sending_loop(self):
        """Sending loop which transmits data from the sending queue to the
        remote socket.

        This method is run as a thread.
        """
        try:
            while not self.done and not self.mc.thread_stopper.is_set():
                try:
                    msg, rawbytes = self.sending_queue.get(block=True,
                                                           timeout=1)

                except queue.Empty:
                    if self.mc.thread_stopper.is_set():
                        self.log.info("Stopping BCP sending thread")
                        self.socket.shutdown(socket.SHUT_RDWR)
                        self.socket.close()
                        self.socket = None
                        return

                    else:
                        continue

                if not rawbytes:
                    self.connection.sendall(('{}\n'.format(msg)).encode('utf-8'))

                else:
                    self.connection.sendall('{}&bytes={}\n'.format(
                        msg, len(rawbytes)).encode('utf-8'))
                    self.connection.sendall(rawbytes)

        except Exception:   # noqa
            exc_type, exc_value, exc_traceback = sys.exc_info()
            lines = traceback.format_exception(exc_type, exc_value,
                                               exc_traceback)
            msg = ''.join(line for line in lines)
            self.mc.crash_queue.put(msg)

            # todo this does not crash mpf-mc

    def process_received_message(self, message):
        """Puts a received BCP message into the receiving queue.

        Args:
            message: The incoming BCP message

        """
        self.log.debug('Received "%s"', message)

        try:
            cmd, kwargs = bcp.decode_command_string(message)
            self.receive_queue.put((cmd, kwargs))
        except ValueError:
            self.log.error("DECODE BCP ERROR. Message: %s", message)
            raise



#!python
#cython: embedsignature=True, language_level=3

from libc.stdio cimport FILE, fopen, fprintf, sprintf, fflush, fclose
from cpython.mem cimport PyMem_Malloc, PyMem_Realloc, PyMem_Free
import cython
import logging
from math import ceil

from mpfmc.core.audio.sdl2 cimport *
from mpfmc.core.audio.gstreamer cimport *
from mpfmc.core.audio.track cimport *
from mpfmc.core.audio.inline cimport in_out_quad
from mpfmc.core.audio.notification_message cimport *
from mpfmc.core.audio.track_sound_loop cimport *


# ---------------------------------------------------------------------------
#    TrackSoundLoop class
# ---------------------------------------------------------------------------
cdef class TrackSoundLoop(Track):
    """
    Live Loop track class
    """

    def __init__(self, object mc, object audio_callback_data, str name, int track_num, int buffer_size,
                 int max_layers=8,
                 float volume=1.0):
        """
        Constructor
        Args:
            mc: The media controller app
            audio_callback_data: The AudioCallbackData struct wrapped in a PyCapsule
            name: The track name
            track_num: The track number
            buffer_size: The length of the track audio buffer in bytes
            max_layers: The maximum number of sounds that can be played simultaneously
                on the track
            volume: The track volume (0.0 to 1.0)
        """
        # IMPORTANT: Call super class init function to allocate track state memory!
        super().__init__(mc, audio_callback_data, name, track_num, buffer_size, volume)

        self.log = logging.getLogger("Track." + str(track_num) + ".TrackSoundLoop." + name)

        SDL_LockAudio()

        # Set track type specific settings
        self.state.mix_callback_function = TrackSoundLoop.mix_playing_sounds

        # Allocate memory for the specific track type state struct (TrackSoundLoopState)
        self.type_state = <TrackSoundLoopState*> PyMem_Malloc(sizeof(TrackSoundLoopState))
        self.state.type_state = <void*>self.type_state

        # Initialize track specific state structures
        self.type_state.players = NULL
        self.type_state.current = NULL

        self._sound_loop_set_counter = 0
        self._active_sound_loop_sets = dict()

        self.log.debug("Created Track %d %s", self.number, self.name)

        SDL_UnlockAudio()

    def __dealloc__(self):
        """Destructor"""

        SDL_LockAudio()

        # Free the specific track type state and other allocated memory
        if self.type_state != NULL:

            if self.type_state.players != NULL:
                iterator = self.type_state.players
                while iterator != NULL:
                    self._delete_player_layers(<SoundLoopSetPlayer*>iterator.data)
                    g_slice_free1(sizeof(SoundLoopSetPlayer), iterator.data)
                    iterator = iterator.next

                g_slist_free(self.type_state.players)
                self.type_state.players = NULL

            PyMem_Free(self.type_state)
            self.type_state = NULL
            if self.state != NULL:
                self.state.type_state = NULL

        SDL_UnlockAudio()

    def __repr__(self):
        return '<Track.{}.SoundLoop.{}>'.format(self.number, self.name)

    @property
    def type(self):
        return "sound_loop"

    @property
    def supports_in_memory_sounds(self):
        """Return whether or not track accepts in-memory sounds"""
        return True

    @property
    def supports_streaming_sounds(self):
        """Return whether or not track accepts streaming sounds"""
        return False

    def stop_all_sounds(self, float fade_out_seconds = 0.0):
        """
        Stops all playing sounds immediately on the track.
        Args:
            fade_out_seconds: The number of seconds to fade out the sounds before stopping
        """
        cdef GSList *iterator
        cdef SoundLoopSetPlayer *player

        iterator = self.type_state.players
        while iterator != NULL:
            player = <SoundLoopSetPlayer*>iterator.data
            # TODO: Stop player (calculate fade out if necessary)

            iterator = iterator.next

    def stop_sound(self, sound not None, fade_out=None):
        """
        Stops all instances of the specified sound immediately on the track. This function is not used
        in the sound loop track, but is provided for compatibility in case the sound player is
        mistakenly used against this track.
        Args:
            sound: The Sound to stop
            fade_out: Optional amount of time (seconds) to fade out before stopping
        """
        pass

    def stop_sound_instance(self, sound_instance not None, fade_out=None):
        """
        Stops the specified sound instance immediately on the track. This function is not used
        in the sound loop track, but is provided for compatibility in case the sound player is
        mistakenly used against this track.
        Args:
            sound_instance: The SoundInstance to stop
            fade_out: Optional amount of time (seconds) to fade out before stopping
        """
        pass

    def stop_sound_looping(self, sound not None):
        """
        Cancels looping for all instances of the specified sound on the track. This function is not used
        in the sound loop track, but is provided for compatibility in case the sound player is
        mistakenly used against this track.
        Args:
            sound: The Sound to stop looping
        """
        pass

    def stop_sound_instance_looping(self, sound_instance not None):
        """
        Stops the specified sound instance on the track after the current loop iteration
        has finished. This function is not used in the sound loop track, but is provided for compatibility
        in case the sound player is mistakenly used against this track.
        Args:
            sound_instance: The Sound to stop
        """
        pass

    def process(self):
        """Processes track messages each tick."""

        # Lock the mutex to ensure no audio data is changed during the playback processing
        # (multi-threaded protection)
        SDL_LockAudio()

        # Process track notification messages
        if self.state.notification_messages != NULL:
            self.state.notification_messages = g_slist_reverse(self.state.notification_messages)
            iterator = self.state.notification_messages
            while iterator != NULL:
                self.process_notification_message(<NotificationMessageContainer*>iterator.data)
                g_slice_free1(sizeof(NotificationMessageContainer), iterator.data)
                iterator = iterator.next

            g_slist_free(self.state.notification_messages)
            self.state.notification_messages = NULL

        # Unlock the mutex since we are done accessing the audio data
        SDL_UnlockAudio()

    cdef process_notification_message(self, NotificationMessageContainer *notification_message):
        """Process a notification message to this track"""

        if notification_message == NULL:
            return

        SDL_LockAudio()

        # Check for track notification messages first (they do not need sound instance information)
        if notification_message.message in (notification_track_stopped, notification_track_paused):
            if notification_message.message == notification_track_stopped:
                # Trigger any events
                if self._events_when_stopped is not None:
                    for event in self._events_when_stopped:
                        self.mc.post_mc_native_event(event, track=self._name)

            elif notification_message.message == notification_track_paused:
                # Trigger any events
                if self._events_when_paused is not None:
                    for event in self._events_when_paused:
                        self.mc.post_mc_native_event(event, track=self._name)
                pass

            SDL_UnlockAudio()
            return

        self.log.debug("Processing notification message %d for sound instance (id: %d)",
                       notification_message.message, notification_message.sound_instance_id)

        if notification_message.message in (notification_sound_loop_set_started, notification_sound_loop_set_stopped,
                                            notification_sound_loop_set_looping):

            if notification_message.data.sound_loop_set.id not in self._active_sound_loop_sets.keys():
                self.log.warning("Received a notification message for a sound loop set (id: %d) "
                                 "that is no longer active. Notification will be discarded.",
                                 notification_message.data.sound_loop_set.id)
                SDL_UnlockAudio()
                return

            sound_loop_set_settings = self._active_sound_loop_sets[notification_message.data.sound_loop_set.id]
            if not sound_loop_set_settings:
                self.log.warning("Could not retrieve settings for a sound loop set (id: %d) "
                                 "Notification will be discarded.",
                                 notification_message.data.sound_loop_set.id)
                SDL_UnlockAudio()
                return

            if notification_message.message == notification_sound_loop_set_started:
                if sound_loop_set_settings['events_when_played'] is not None:
                    for event in sound_loop_set_settings['events_when_played']:
                        self.mc.post_mc_native_event(event)

            elif notification_message.message == notification_sound_loop_set_stopped:
                if sound_loop_set_settings['events_when_stopped'] is not None:
                    for event in sound_loop_set_settings['events_when_stopped']:
                        self.mc.post_mc_native_event(event)

                self.log.debug("Removing sound_loop_set settings %d from active list", notification_message.data.sound_loop_set.id)
                del self._active_sound_loop_sets[notification_message.data.sound_loop_set.id]

                # Remove and delete sound loop set player used to play this set
                player = <SoundLoopSetPlayer*>notification_message.data.sound_loop_set.player
                if player != NULL:
                    self._delete_player(player)
                    self.type_state.players = g_slist_remove(self.type_state.players, player)

            elif notification_message.message == notification_sound_loop_set_looping:
                if sound_loop_set_settings['events_when_looping'] is not None:
                    for event in sound_loop_set_settings['events_when_looping']:
                        self.mc.post_mc_native_event(event)

        elif notification_message.message == notification_sound_marker:

            if notification_message.sound_id not in self.mc.sounds_by_id.keys():
                self.log.warning("Received a notification message for a sound (id: %d) "
                                 "that does not exist. Notification will be discarded.",
                                 notification_message.sound_id)
                SDL_UnlockAudio()
                return

            # In a sound loop track, sound instances are not created when playing a loop sound.  The sound
            # asset is used directly and therefore we need to reference the sound asset by sound_id to
            # retrieve the marker events to send.
            sound = self.mc.sounds_by_id[notification_message.sound_id]

            # Send marker event(s)
            if sound and 0 <= notification_message.data.marker.id < sound.marker_count:
                if sound.markers[notification_message.data.marker.id]['events'] is not None:
                    for event in sound.markers[notification_message.data.marker.id]['events']:
                        self.mc.post_mc_native_event(event,
                                                     sound_instance=sound,
                                                     marker_id=notification_message.data.marker.id)

        SDL_UnlockAudio()

    def play_sound_loop_set(self, dict sound_loop_set not None, str context=None, dict player_settings=None):
        """
        Immediately play a sound loop set.

        Args:
            sound_loop_set: The sound_loop_set asset object to play.
            context: The context from which the sound loop set is played.
            player_settings: Settings to use for playback
        """
        cdef SoundLoopSetPlayer *player
        cdef SoundLoopSetPlayer *existing_player
        cdef SoundLoopLayerSettings *layer
        cdef int bytes_per_sample_frame
        cdef bint player_already_playing = False

        self.log.debug("play_sound_loop_set - Preparing sound_loop_set '%s' for playback.", sound_loop_set)

        if player_settings is None:
            player_settings = dict()

        # Determine settings (override sound loop set with player settings)
        player_settings.setdefault('fade_in', sound_loop_set['fade_in'])
        player_settings.setdefault('fade_out', sound_loop_set['fade_out'])
        player_settings.setdefault('start_at', 0)
        player_settings.setdefault('events_when_played', sound_loop_set['events_when_played'])
        player_settings.setdefault('events_when_stopped', sound_loop_set['events_when_stopped'])
        player_settings.setdefault('events_when_looping', sound_loop_set['events_when_looping'])
        player_settings.setdefault('mode_end_action', sound_loop_set['mode_end_action'])
        player_settings.setdefault('tempo', sound_loop_set['tempo'])
        player_settings.setdefault('timing', 'loop_end')
        player_settings.setdefault('interval', 1)
        player_settings.setdefault('synchronize', False)
        player_settings['sound'] = sound_loop_set['sound']
        player_settings['context'] = context

        # print("play_sound_loop_set - player_settings:", player_settings)

        if player_settings['sound'] not in self.mc.sounds:
            self.log.error("play_sound_loop_set - sound %s in sound_loop_set could "
                           "not be located, could not play loop set", player_settings['sound'])
            return

        if not self.mc.sounds[player_settings['sound']].loaded:
            self.log.error("play_sound_loop_set - sound %s in sound_loop_set was "
                           "not loaded, could not play loop set", player_settings['sound'])
            return

        SDL_LockAudio()

        # Create new player
        player = <SoundLoopSetPlayer*>g_slice_alloc0(sizeof(SoundLoopSetPlayer))
        self._initialize_player(player)

        self._sound_loop_set_counter += 1

        # Assign sound loop settings
        master_layer_settings = {
            "sound_loop_set_id": self._sound_loop_set_counter,
            "sound": sound_loop_set['sound'],
            "volume": sound_loop_set['volume'],
            "initial_state": "play"
        }

        self._apply_layer_settings(&player.master_sound_layer, master_layer_settings)

        # Determine master sound length
        if player.master_sound_layer.sound != NULL:
            player.length = <Uint32>player.master_sound_layer.sound.data.memory.size
            player.tempo = player_settings['tempo']
        else:
            player.length = 0
            player.sample_pos = 0
            player.status = player_idle
            g_slice_free1(sizeof(SoundLoopSetPlayer), player)
            SDL_UnlockAudio()
            return

        # Setup sound loop set layers
        for layer_settings in sound_loop_set['layers']:
            layer = _create_sound_loop_layer_settings()
            self._apply_layer_settings(layer, layer_settings)

            # Layer fading is only set by events using the sound_loop_player
            layer.fade_in_steps = 0
            layer.fade_out_steps = 0
            layer.fade_steps_remaining = 0

            # Append layer
            player.layers = g_slist_append(player.layers, layer)

        # Calculate fading (done at control rate; need to calculate the number of steps over which to fade in/out)
        player.master_sound_layer.fade_in_steps = player_settings['fade_in'] * self.state.callback_data.seconds_to_bytes_factor // self.state.callback_data.bytes_per_control_point
        player.master_sound_layer.fade_out_steps = player_settings['fade_out'] * self.state.callback_data.seconds_to_bytes_factor // self.state.callback_data.bytes_per_control_point
        player.master_sound_layer.fade_steps_remaining = player.master_sound_layer.fade_in_steps

        # Determine if there are any other active players
        if self.type_state.players != NULL:
            iterator = self.type_state.players
            while iterator != NULL:
                existing_player = <SoundLoopSetPlayer*>iterator.data
                if existing_player != NULL and existing_player.status != player_idle:
                    player_already_playing = True
                    break
                iterator = iterator.next

        if player_already_playing and self.type_state.current != NULL:

            # Determine if playing immediately or queuing until next loop
            if player_settings['timing'] == 'loop_end':
                # Queue until the end of the current loop
                player.status = player_delayed

                # Ensure sample position starts on a sample frame boundary (audio distortion may occur if starting
                # in the middle of a sample frame)
                bytes_per_sample_frame = self.state.callback_data.bytes_per_sample * self.state.callback_data.channels
                player.sample_pos = bytes_per_sample_frame * ceil(player.sample_pos / bytes_per_sample_frame)

                self.type_state.current.stop_loop_samples_remaining = self.type_state.current.length - self.type_state.current.sample_pos
                player.start_delay_samples_remaining = self.type_state.current.stop_loop_samples_remaining

            elif player_settings['timing'] == 'next_time_interval':
                # Set currently playing sample to end at the next specified time interval multiple
                player.status = player_delayed

                self.type_state.current.stop_loop_samples_remaining = self._round_sample_pos_up_to_interval(
                    self.type_state.current.sample_pos,
                    <Uint32> (self.state.callback_data.seconds_to_bytes_factor * player_settings['interval']),
                    self.state.callback_data.bytes_per_sample * self.state.callback_data.channels) - self.type_state.current.sample_pos

                player.start_delay_samples_remaining = self.type_state.current.stop_loop_samples_remaining

            elif player_settings['timing'] == 'next_beat_interval':
                # Set currently playing sample to end at the next specified beat interval multiple
                player.status = player_delayed

                self.type_state.current.stop_loop_samples_remaining = self._round_sample_pos_up_to_interval(
                    self.type_state.current.sample_pos,
                    <Uint32> (self.state.callback_data.seconds_to_bytes_factor * (
                            player_settings['interval'] * 60.0 / self.type_state.current.tempo)),
                    self.state.callback_data.bytes_per_sample * self.state.callback_data.channels) - self.type_state.current.sample_pos

                player.start_delay_samples_remaining = self.type_state.current.stop_loop_samples_remaining

            else:
                # 'timing' == 'now'
                player.status = player_playing

                # If no fade is set, use a quick fade-in
                # (avoids pops & clicks)
                if player.master_sound_layer.fade_steps_remaining == 0:
                    player.master_sound_layer.fade_in_steps = self.state.callback_data.quick_fade_steps
                    player.master_sound_layer.fade_steps_remaining = player.master_sound_layer.fade_in_steps

                if player.master_sound_layer.fade_steps_remaining > 0:
                    player.status = player_fading_in
                    self._fade_out_all_players(player.master_sound_layer.fade_in_steps)
                    # TODO: Should layers fade out as well? (Probably)

            # Cancel and remove all existing delayed players
            self._cancel_all_delayed_players()

            if player_settings['synchronize']:
                player.sample_pos = self.type_state.current.sample_pos + self.type_state.current.stop_loop_samples_remaining

                # Adjust sample position to within sample boundaries of synchronized loop
                while player.sample_pos > self.type_state.current.length:
                    player.sample_pos -= self.type_state.current.length
            else:
                player.sample_pos = player_settings['start_at'] * self.state.callback_data.seconds_to_bytes_factor

        else:
            # 'timing' value is ignored when there are no other sound loops playing (playback
            # begins right away)
            player.master_sound_layer.fade_steps_remaining = 0
            player.sample_pos = player_settings['start_at'] * self.state.callback_data.seconds_to_bytes_factor
            player.status = player_playing

        # Ensure sample position starts on a sample frame boundary (audio distortion may occur if starting
        # in the middle of a sample frame)
        bytes_per_sample_frame = self.state.callback_data.bytes_per_sample * self.state.callback_data.channels
        player.sample_pos = bytes_per_sample_frame * ceil(player.sample_pos / bytes_per_sample_frame)

        # Adjust new player playing position to ensure it is within the sample
        while player.sample_pos >= player.length:
            player.sample_pos -= player.length

        # Save current sound loop set so it can be referred to again while it is active (event notifications)
        self._active_sound_loop_sets[self._sound_loop_set_counter] = player_settings

        # Add new player to the player list
        self.type_state.players = g_slist_prepend(self.type_state.players, player)

        # Send sound_loop_set started notification (if not pending/queued)
        if player.status not in (player_idle, player_delayed):
            self.type_state.current = player
            send_sound_loop_set_started_notification(player.master_sound_layer.sound_loop_set_id,
                                                     player.master_sound_layer.sound_id,
                                                     player,
                                                     self.state)

        # print("play_sound_loop_set - status: ", self.get_status())

        SDL_UnlockAudio()

    cdef _apply_layer_settings(self, SoundLoopLayerSettings *layer, dict layer_settings):
        cdef SoundFile sound_container

        if 'initial_state' not in layer_settings or layer_settings['initial_state'] == 'play':
            layer.status = layer_playing
        else:
            layer.status = layer_stopped

        if 'sound_loop_set_id' in layer_settings:
            layer.sound_loop_set_id = layer_settings['sound_loop_set_id']
        else:
            layer.sound_loop_set_id = 0

        # Set layer sound
        sound = self.mc.sounds[layer_settings['sound']]

        # TODO: What to do when sound is not loaded?
        # TODO: Perhaps load sounds and delay play until loaded?

        sound_container = sound.container
        layer.sound = cython.address(sound_container.sample)
        layer.sound_id = sound.id

        # By default, all layers will continue to loop when played
        layer.looping = True

        # Layer volume (use layer settings or sound setting if None)
        if layer_settings['volume']:
            layer.volume = <Uint8>(layer_settings['volume'] * SDL_MIX_MAXVOLUME)
        else:
            layer.volume = <Uint8>(sound.volume * SDL_MIX_MAXVOLUME)

        # Markers (copy from source sound)
        layer.marker_count = sound.marker_count

        if layer.marker_count > 0:
            layer.markers = g_array_new(False, False, sizeof(guint))
            g_array_set_size(layer.markers, sound.marker_count)
            for index in range(sound.marker_count):
                g_array_insert_val_uint(layer.markers,
                                        index,
                                        <guint>(sound.markers[index]['time'] * self.state.callback_data.seconds_to_bytes_factor))

    def stop_current_sound_loop_set(self, fade_out=None):
        """
        Stops the currently playing sound loop set.
        Args:
            fade_out: The number of seconds over which to fade out the currently playing
                      sound loop set before stopping.
        """
        SDL_LockAudio()

        self.log.debug("Stopping current sound loop set")

        if self.type_state.current == NULL or self.type_state.current.status not in (player_playing, player_fading_in, player_fading_out):
            self.log.info("Unable to stop sound loop set - no sound loop set is currently playing.")
            SDL_UnlockAudio()
            return

        # Calculate new fade out if specified (overriding current setting)
        if fade_out:
            self.type_state.current.master_sound_layer.fade_out_steps = fade_out * self.state.callback_data.seconds_to_bytes_factor // self.state.callback_data.bytes_per_control_point

        # If no fade out is specified, perform quick fade out
        if self.type_state.current.master_sound_layer.fade_out_steps == 0:
            self.type_state.current.master_sound_layer.fade_out_steps = self.state.callback_data.quick_fade_steps

        self.type_state.current.master_sound_layer.fade_steps_remaining = self.type_state.current.master_sound_layer.fade_out_steps
        self.type_state.current.status = player_fading_out

        # Need to check if the next sound player has a loop set queued for playback and if so, cancel it
        self._cancel_all_delayed_players()

        # Do we need to shorted the current fade-out?
        self._fade_out_all_players(self.type_state.current.master_sound_layer.fade_steps_remaining)

        SDL_UnlockAudio()

    def jump_to_time_current_sound_loop_set(self, time=0.0):
        """Immediately jumps the loop playback position to the specified time."""

        SDL_LockAudio()

        self.log.debug("Jumping to %f seconds playback position in current sound loop set", time)

        if self.type_state.current == NULL or self.type_state.current.status not in (player_playing, player_fading_in, player_fading_out):
            self.log.info("Unable to jump to specified playback position - "
                          "no sound loop set is currently playing.")
            SDL_UnlockAudio()
            return

        # Set new playback position (make sure it is within the current sample length)
        self.type_state.current.sample_pos = time * self.state.callback_data.seconds_to_bytes_factor
        while self.type_state.current.sample_pos >= self.type_state.current.length:
            self.type_state.current.sample_pos -= self.type_state.current.length

        SDL_UnlockAudio()

    def stop_looping_current_sound_loop_set(self):
        """
        Stops the currently playing sound loop set from continuing to loop (it will stop
        after the current loop iteration).
        """
        SDL_LockAudio()

        self.log.debug("Stopping looping current sound loop set")

        if self.type_state.current == NULL or self.type_state.current.status not in (player_playing, player_fading_in, player_fading_out):
            self.log.info("Unable to stop looping sound loop set - no sound loop set is currently playing.")
            SDL_UnlockAudio()
            return

        self.type_state.current.stop_loop_samples_remaining = self.type_state.current.length - self.type_state.current.sample_pos

        SDL_UnlockAudio()

    def play_layer(self, int layer, float fade_in=0.0, str timing="loop_end", volume=None):
        """
        Plays the specified layer number in the currently playing sound loop set.

        Args:
            layer: The layer number to play (1-based array position).
            fade_in: The number of seconds over which to fade in the layer.
            timing: Flag indicating whether the layer should be played immediately ('now')
                    or queued until the next loop of the master layer loop ('loop_end').
            volume: The layer volume (overrides sound loop set setting if not None)
        """
        cdef SoundLoopLayerSettings *layer_settings

        if layer < 1:
            self.log.warning("Illegal layer value in call to play_layer (must be > 0).")
            return

        SDL_LockAudio()

        self.log.debug("play_layer - Play layer %d of the currently playing sound_loop_set (fade-in = %f sec).",
                       layer, fade_in)

        if self.type_state.current == NULL or self.type_state.current.status not in (player_playing, player_fading_in, player_fading_out):
            self.log.info("Unable to play layer - no sound loop set is currently playing.")
            SDL_UnlockAudio()
            return

        # Retrieve the layer
        if self.type_state.current.layers == NULL:
            self.log.info("There are no layers defined in the current sound loop set: play_layers has no effect")
            SDL_UnlockAudio()
            return

        layer_settings = <SoundLoopLayerSettings*>g_slist_nth_data(self.type_state.current.layers, layer - 1)
        if layer_settings == NULL:
            self.log.info("The specified layer could not be found in the current sound loop set: play_layers has no effect")
            SDL_UnlockAudio()
            return

        if layer_settings.status != layer_stopped:
            self.log.info("The current sound loop set layer is already playing: play_layers has no effect")
            SDL_UnlockAudio()
            return

        # Layer volume (use layer settings or sound setting if None)
        if volume:
            layer_settings.volume = <Uint8>(volume * SDL_MIX_MAXVOLUME)

        # Calculate fading (done at control rate; need to calculate the number of steps over which to fade in/out)
        if fade_in > 0.0:
            layer_settings.fade_in_steps = <Uint32>(fade_in * self.state.callback_data.seconds_to_bytes_factor) // self.state.callback_data.bytes_per_control_point
            layer_settings.fade_steps_remaining = layer_settings.fade_in_steps

        if timing == 'loop_end':
            layer_settings.status = layer_queued

        elif timing == 'now':
            if layer_settings.fade_in_steps > 0:
                layer_settings.status = layer_fading_in
            else:
                layer_settings.status = layer_playing
        else:
            self.log.error("Unknown timing value specified: %s", timing)
            SDL_UnlockAudio()
            return

        layer_settings.looping = True

        SDL_UnlockAudio()

    def stop_layer(self, int layer, float fade_out=0.0):
        """
        Stops the specified layer number in the currently playing sound loop set.

        Args:
            layer: The layer number to stop (1-based array position).
            fade_out: The number of seconds over which to fade out the layer before stopping.
        """
        cdef SoundLoopLayerSettings *layer_settings

        if layer < 1:
            self.log.warning("Illegal layer value in call to stop_layer (must be > 0).")
            return

        SDL_LockAudio()

        self.log.debug("stop_layer - Stop layer %d of the currently playing sound_loop_set (fade-out = %f sec).",
                       layer, fade_out)

        if self.type_state.current == NULL or self.type_state.current.status not in (player_playing, player_fading_in, player_fading_out):
            self.log.info("Unable to stop layer - no sound loop set is currently playing.")
            SDL_UnlockAudio()
            return

        # Retrieve the layer
        if self.type_state.current.layers == NULL:
            self.log.info("There are no layers defined in the current sound loop set: stop_layers has no effect")
            SDL_UnlockAudio()
            return

        layer_settings = <SoundLoopLayerSettings*>g_slist_nth_data(self.type_state.current.layers, layer - 1)
        if layer_settings == NULL:
            self.log.info("The specified layer could not be found in the current sound loop set: stop_layers has no effect")
            SDL_UnlockAudio()
            return

        if layer_settings.status == layer_stopped:
            self.log.info("The current sound loop set layer is already stopped: stop_layers has no effect")
            SDL_UnlockAudio()
            return

        # Calculate fading (done at control rate; need to calculate the number of steps over which to fade in/out)
        if fade_out > 0.0:
            layer_settings.fade_out_steps = <Uint32>(fade_out * self.state.callback_data.seconds_to_bytes_factor) // self.state.callback_data.bytes_per_control_point
            layer_settings.fade_steps_remaining = layer_settings.fade_out_steps
            layer_settings.status = layer_fading_out
        else:
            # TODO: Could perform a quick fade out rather than an abrupt stop
            layer_settings.status = layer_stopped

        SDL_UnlockAudio()

    def stop_looping_layer(self, int layer):
        """
        Stops the specified layer number in the currently playing sound loop set after
        the currently playing loop finishes.

        Args:
            layer: The layer number to stop looping (1-based array position).
        """
        cdef SoundLoopLayerSettings *layer_settings

        if layer < 1:
            self.log.warning("Illegal layer value in call to stop_looping_layer (must be > 0).")
            return

        SDL_LockAudio()

        self.log.debug("stop_looping_layer - Stop looping layer %d of the currently playing sound_loop_set.", layer)

        # Retrieve the layer
        if self.type_state.current.layers == NULL:
            self.log.info("There are no layers defined in the current sound loop set: "
                          "stop_looping_layers has no effect")
            SDL_UnlockAudio()
            return

        layer_settings = <SoundLoopLayerSettings*>g_slist_nth_data(self.type_state.current.layers, layer - 1)
        if layer_settings == NULL:
            self.log.info("The specified layer could not be found in the current sound loop set: "
                          "stop_looping_layers has no effect")
            SDL_UnlockAudio()
            return

        if layer_settings.status == layer_stopped:
            self.log.info("The current sound loop set layer is already stopped: "
                          "stop_looping_layers has no effect")
            SDL_UnlockAudio()
            return

        layer_settings.looping = False

        SDL_UnlockAudio()

    cdef _initialize_player(self, SoundLoopSetPlayer *player):
        """Initializes a SoundLoopSetPlayer struct."""
        if player != NULL:
            player.layers = NULL
            player.master_sound_layer.markers = NULL
            player.status = player_idle
            player.length = 0
            player.master_sound_layer.status = layer_playing
            player.master_sound_layer.sound = NULL
            player.master_sound_layer.volume = 0
            player.master_sound_layer.sound_loop_set_id = 0
            player.master_sound_layer.sound_id = 0
            player.master_sound_layer.fade_in_steps = 0
            player.master_sound_layer.fade_out_steps = 0
            player.master_sound_layer.fade_steps_remaining = 0
            player.master_sound_layer.looping = True
            player.master_sound_layer.marker_count = 0
            player.sample_pos = 0
            player.start_delay_samples_remaining = 0
            player.stop_loop_samples_remaining = do_not_stop_loop

    cdef _delete_player(self, SoundLoopSetPlayer *player):
        """Delete player (frees any allocated memory, including in layers)"""
        if player != NULL:
            self._delete_player_layers(player)
            g_slice_free1(sizeof(SoundLoopSetPlayer), player)

    cdef _delete_player_layers(self, SoundLoopSetPlayer *player):
        """Delete (free memory) for sound loop set player layers."""
        if player != NULL:
            if player.master_sound_layer.markers != NULL:
                g_array_free(player.master_sound_layer.markers, True)

            if player.layers != NULL:
                iterator = player.layers
                while iterator != NULL:
                    layer = <SoundLoopLayerSettings*>iterator.data
                    if layer.markers != NULL:
                        g_array_free(layer.markers, True)
                    g_slice_free1(sizeof(SoundLoopLayerSettings), layer)
                    iterator = iterator.next

                g_slist_free(player.layers)

    cdef _cancel_all_delayed_players(self):
        """Cancels and removes all delayed (pending) players."""
        cdef SoundLoopSetPlayer *player
        cdef GSList *iterator = self.type_state.players

        # Loop over players
        while iterator != NULL:
            player = <SoundLoopSetPlayer*>iterator.data
            iterator = iterator.next

            # Remove player if it currently has a delayed status
            if player != NULL and player.status == player_delayed:
                self.type_state.players = g_slist_remove(self.type_state.players, player)

                # Remove sound loop set from list of active sound loop sets
                if player.master_sound_layer.sound_id in self._active_sound_loop_sets.keys():
                    del self._active_sound_loop_sets[player.master_sound_layer.sound_id]

                # Clean up allocated player memory
                self._delete_player_layers(player)
                g_slice_free1(sizeof(SoundLoopSetPlayer), player)

    cdef _fade_out_all_players(self, Uint32 fade_steps):
        """Fades out all currently playing players."""
        cdef SoundLoopSetPlayer *player
        cdef GSList *iterator = self.type_state.players

        # Loop over players
        while iterator != NULL:
            player = <SoundLoopSetPlayer*>iterator.data
            iterator = iterator.next

            if player == NULL:
                continue

            elif player.status == player_playing:
                player.status = player_fading_out
                player.master_sound_layer.fade_out_steps = fade_steps
                player.master_sound_layer.fade_steps_remaining = fade_steps

            elif player.status == player_fading_out:
                # The existing fade will only be adjusted if it has more steps remaining than specified here
                if player.master_sound_layer.fade_steps_remaining > fade_steps:
                    player.master_sound_layer.fade_out_steps = <Uint32>round(fade_steps * player.master_sound_layer.fade_out_steps / player.master_sound_layer.fade_steps_remaining)
                    player.master_sound_layer.fade_steps_remaining = fade_steps

            elif player.status == player_fading_in:
                # Keep the same volume level to start as the current fade in
                player.master_sound_layer.fade_out_steps = <Uint32>round(fade_steps / (1.0 - (player.master_sound_layer.fade_steps_remaining / player.master_sound_layer.fade_in_steps)))
                player.master_sound_layer.fade_steps_remaining = fade_steps
                player.status = player_fading_out

    def get_status(self):
        """
        Get the current track status (status of all sound loop players on the track).
        Used for debugging and testing.
        Returns:
            A list of status dictionaries containing the current settings for each
            sound loop player.
        """
        cdef SoundLoopSetPlayer *player
        cdef SoundLoopLayerSettings *layer
        cdef int index = 0
        cdef GSList *iterator

        SDL_LockAudio()
        status = []

        iterator = self.type_state.players
        while iterator != NULL:
            player = <SoundLoopSetPlayer*>iterator.data
            if player == NULL:
                status.append({
                    "status": TrackSoundLoop.player_status_to_text(player_idle),
                    "length": 0
                })
            else:
                layers = []
                layer = <SoundLoopLayerSettings*>g_slist_nth_data(player.layers, index)
                while layer != NULL:
                    layers.append({
                        "status": TrackSoundLoop.layer_status_to_text(layer.status),
                        "sound_id": layer.sound_id,
                        "sound_length": layer.sound.data.memory.size,
                        "volume": layer.volume,
                        "fade_in_steps": layer.fade_in_steps,
                        "fade_out_steps": layer.fade_out_steps,
                        "fade_steps_remaining": layer.fade_steps_remaining,
                        "looping": layer.looping,
                        "marker_count": layer.marker_count,
                    })
                    index += 1
                    layer = <SoundLoopLayerSettings*>g_slist_nth_data(player.layers, index)

                status.append({
                    "status": TrackSoundLoop.player_status_to_text(<int>player.status),
                    "length": player.length,
                    "sound_id": player.master_sound_layer.sound_id,
                    "volume": player.master_sound_layer.volume,
                    "tempo": player.tempo,
                    "layers": layers,
                    "sample_pos": player.sample_pos,
                    "stop_loop_samples_remaining": "DO NOT STOP LOOP" if player.stop_loop_samples_remaining == do_not_stop_loop else player.stop_loop_samples_remaining,
                    "start_delay_samples_remaining": player.start_delay_samples_remaining,
                    "fade_in_steps": player.master_sound_layer.fade_in_steps,
                    "fade_out_steps": player.master_sound_layer.fade_out_steps,
                    "fade_steps_remaining": player.master_sound_layer.fade_steps_remaining,
                })

            iterator = iterator.next

        SDL_UnlockAudio()

        return status

    @staticmethod
    def player_status_to_text(int status):
        """
        Converts a sound loop player status value into an equivalent text string.  Used for testing
        purposes only.
        Args:
            status: Integer sound loop player status value

        Returns:
            string containing the equivalent status text
        """
        status_values = {
            player_idle: "idle",
            player_delayed: "delayed",
            player_playing: "playing",
            player_fading_in: "fading in",
            player_fading_out: "fading out",
        }

        try:
            return status_values.get(status)
        except KeyError:
            return "unknown"

    @staticmethod
    def layer_status_to_text(int status):
        """
        Converts a sound loop set layer status value into an equivalent text string.  Used for testing
        purposes only.
        Args:
            status: Integer sound loop set layer status value

        Returns:
            string containing the equivalent status text
        """
        status_values = {
            layer_stopped: "stopped",
            layer_queued: "queued",
            layer_playing: "playing",
            layer_fading_in: "fading in",
            layer_fading_out: "fading out",
        }

        try:
            return status_values.get(status)
        except KeyError:
            return "unknown"

    cdef inline Uint32 _round_sample_pos_up_to_interval(self, Uint32 sample_pos, Uint32 interval, int bytes_per_sample_frame):
        """
        Rounds sample position up to the nearest specified interval.
        
        Args:
            sample_pos: The sample position value to round up. 
            interval: The interval to round up to.
            bytes_per_sample_frame: The number of bytes per sample frame (a frame includes a single sample
                for each channel in the audio).

        Returns:
            Sample position rounded up to the nearest specified interval.
            
        Notes:
            The number of audio channels is important in this function so we do not split a sample frame in
            the middle of its bytes (would cause unwanted audio artifacts). Buffers are only split on
            a whole sample frame boundary.
        """
        return bytes_per_sample_frame * ceil((interval * ceil(sample_pos / interval)) / bytes_per_sample_frame)


    # ---------------------------------------------------------------------------
    #    Static C functions designed to be called from the static audio callback
    #    function (these functions do not use the GIL).
    #
    #    Note: Because these functions are only called from the audio callback
    #    function, we do not need to lock and unlock the mutex in these functions
    #    (locking/unlocking of the mutex is already performed in the audio
    #    callback function.
    # ---------------------------------------------------------------------------

    @staticmethod
    cdef void mix_playing_sounds(TrackState *track, Uint32 buffer_length, AudioCallbackData *callback_data) nogil:
        """
        Mixes any sounds that are playing on the specified standard track into the specified audio buffer.
        Args:
            track: A pointer to the TrackState data structure for the track
            buffer_length: The length of the output buffer (in bytes)
            callback_data: The audio callback data structure
        Notes:
            Notification messages are generated.
        """
        cdef TrackState *target_track
        cdef TrackSoundLoopState *sound_loop_track
        cdef bint first_layer = True
        cdef GSList *iterator
        cdef SoundLoopSetPlayer *player

        if track == NULL or track.type_state == NULL:
            return

        sound_loop_track = <TrackSoundLoopState*>track.type_state

        # Setup local variables
        cdef Uint32 track_buffer_pos

        # Loop over the current sound loop players
        iterator = sound_loop_track.players
        while iterator != NULL:
            player = <SoundLoopSetPlayer*>iterator.data
            iterator = iterator.next

            # If the current player is idle, the track is not active so there is nothing to do for this player
            if player == NULL or player.status == player_idle:
                continue

            # Process the current loop player (fill the track buffer with samples)
            track_buffer_pos = 0

            # Handle any player delay
            if player.status == player_delayed:

                # Determine if player should start (delay elapses during this buffer)
                if player.start_delay_samples_remaining < buffer_length:

                    # Start player playing
                    track_buffer_pos = player.start_delay_samples_remaining
                    player.start_delay_samples_remaining = 0
                    player.status = player_playing

                    # This player now becomes the current player
                    sound_loop_track.current = player

                    # Send started notification
                    send_sound_loop_set_started_notification(player.master_sound_layer.sound_loop_set_id,
                                                             player.master_sound_layer.sound_id,
                                                             player,
                                                             track)

                else:
                    # Continue to delay until a future buffer
                    player.start_delay_samples_remaining -= buffer_length
                    track_buffer_pos = buffer_length

            # Loop over track buffer, filling samples for player (if applicable)
            while track_buffer_pos < buffer_length and player.status \
                    in (player_playing, player_fading_in, player_fading_out):

                # Set flag indicating there is at least some activity on the track (it is active)
                track.active = True

                track_buffer_pos = get_player_sound_samples(track, player, buffer_length,
                                                            track_buffer_pos, callback_data)

                # Determine if the player has finished
                if player.status == player_idle:

                    # Player has stopped and is now idle, if it is set as the current one, set current to NULL
                    if sound_loop_track.current == player:
                        sound_loop_track.current = NULL

                    # Send stopped notification
                    send_sound_loop_set_stopped_notification(player.master_sound_layer.sound_loop_set_id,
                                                             player.master_sound_layer.sound_id,
                                                             player,
                                                             track)

                # Determine if sound is looping
                elif player.sample_pos >= player.length:

                    # Send looping notification
                    send_sound_loop_set_looping_notification(player.master_sound_layer.sound_loop_set_id,
                                                             player.master_sound_layer.sound_id,
                                                             player,
                                                             track)
                    player.sample_pos = 0

cdef Uint32 get_player_sound_samples(TrackState *track, SoundLoopSetPlayer *player,
                                     Uint32 buffer_length, Uint32 track_buffer_pos,
                                     AudioCallbackData *callback_data) nogil:
    """
    Retrieves and copies samples from the source sound to fill the track buffer.
     
    Args:
        track: 
        player: 
        buffer_length: 
        track_buffer_pos: 
        callback_data: 

    Returns:
        The updated track buffer position.
        
    Notes:
        The main loop fills the output buffer in chunks that are as large as possible, up to a maximum
        of the buffer length.  The chunks will be smaller during fading as each chunk will be at
        the same volume level (volume is updated at control rate and not audio rate). Also, if a
        sound finishes the chunk will only be big enough for the tail end of the sound.  This means
        it is possible to have a chunk that is only one sample long.
    """

    cdef SoundLoopLayerSettings *layer
    cdef GSList *layer_iterator
    cdef Uint8 player_volume, layer_volume
    cdef Uint32 layer_sample_pos, layer_bytes_remaining, layer_chunk_bytes, current_chunk_bytes, temp_chunk_bytes
    cdef Uint32 layer_track_buffer_pos_offset
    cdef Uint32 buffer_bytes_remaining = buffer_length - track_buffer_pos

    # Loop over the output buffer
    while buffer_bytes_remaining > 0:

        # Mix current player into track buffer.  Processing is done at control rate during player
        # fading and using full buffer when no fading is occurring.

        # Calculate the size and volume of chunk (handle fading)
        if player.status == player_fading_in:
            current_chunk_bytes = min(buffer_bytes_remaining,
                                      player.length - player.sample_pos,
                                      callback_data.bytes_per_control_point)
            player_volume = <Uint8> (in_out_quad((player.master_sound_layer.fade_in_steps - player.master_sound_layer.fade_steps_remaining) / player.master_sound_layer.fade_in_steps) * SDL_MIX_MAXVOLUME)
            player.master_sound_layer.fade_steps_remaining -= 1
            if player.master_sound_layer.fade_steps_remaining == 0:
                player.status = player_playing

        elif player.status == player_fading_out:
            current_chunk_bytes = min(buffer_bytes_remaining,
                                      player.length - player.sample_pos,
                                      callback_data.bytes_per_control_point)
            player_volume = <Uint8> (in_out_quad(player.master_sound_layer.fade_steps_remaining / player.master_sound_layer.fade_out_steps) * SDL_MIX_MAXVOLUME)
            player.master_sound_layer.fade_steps_remaining -= 1
            if player.master_sound_layer.fade_steps_remaining == 0:
                player.status = player_idle
        else:
            current_chunk_bytes = min(buffer_bytes_remaining, player.length - player.sample_pos)
            player_volume = SDL_MIX_MAXVOLUME

        # Adjust chunk size if the loop is set to stop early (if not the chunk size will remain unchanged)
        current_chunk_bytes = min(current_chunk_bytes, player.stop_loop_samples_remaining)

        if player.status == player_idle or current_chunk_bytes == 0:
            return buffer_length

        # Get master layer samples
        SDL_MixAudioFormat(track.buffer + track_buffer_pos,
                           <Uint8*>player.master_sound_layer.sound.data.memory.data + player.sample_pos,
                           track.callback_data.format,
                           current_chunk_bytes,
                           player_volume * player.master_sound_layer.volume // SDL_MIX_MAXVOLUME)

        # Process markers (do any markers fall in the current chunk?)
        # Note: the current sample position has already been incremented when the sample data was received so
        # we need to look backwards from the current position to determine if marker falls in chunk window.
        for marker_id in range(player.master_sound_layer.marker_count):
            if player.sample_pos - current_chunk_bytes <= g_array_index_uint(player.master_sound_layer.markers, marker_id) < player.sample_pos:
                # Marker is in window, send notification
                send_sound_marker_notification(0,
                                               player.master_sound_layer.sound_id,
                                               player.master_sound_layer.sound_loop_set_id,
                                               track,
                                               marker_id)

        # Now mix any additional loop layers
        layer_iterator = player.layers
        while layer_iterator != NULL:
            layer = <SoundLoopLayerSettings*>layer_iterator.data

            if player.sample_pos == 0:

                if layer.status == layer_queued:
                    layer.status = layer_playing

                if not layer.looping:
                    layer.status = layer_stopped

            if layer.status in (layer_playing, layer_fading_in, layer_fading_out):
                layer_track_buffer_pos_offset = 0
                layer_sample_pos = player.sample_pos
                layer_bytes_remaining = current_chunk_bytes

                while layer_bytes_remaining > 0:
                    if layer.status == layer_playing:
                        layer_chunk_bytes = layer_bytes_remaining
                        layer_volume = layer.volume
                    else:
                        # Calculate layer volume (handle fading)
                        layer_chunk_bytes = min(callback_data.bytes_per_control_point, layer_bytes_remaining)
                        if layer.status == layer_fading_in:
                            layer_volume = <Uint8> (in_out_quad((layer.fade_in_steps - layer.fade_steps_remaining) / layer.fade_in_steps) * layer.volume)
                            layer.fade_steps_remaining -= 1
                            if layer.fade_steps_remaining == 0:
                                layer.status = layer_playing
                        elif layer.status == layer_fading_out:
                            layer_volume = <Uint8> (in_out_quad(layer.fade_steps_remaining / layer.fade_out_steps) * layer.volume)
                            layer.fade_steps_remaining -= 1
                            if layer.fade_steps_remaining == 0:
                                layer.status = layer_stopped
                                layer_bytes_remaining = 0

                    if layer_sample_pos < layer.sound.data.memory.size:
                        SDL_MixAudioFormat(track.buffer + track_buffer_pos + layer_track_buffer_pos_offset,
                                           <Uint8*>layer.sound.data.memory.data + layer_sample_pos,
                                           callback_data.format,
                                           min(layer_chunk_bytes, layer.sound.data.memory.size - layer_sample_pos),
                                           player_volume * layer_volume // SDL_MIX_MAXVOLUME)

                        # Process markers (do any markers fall in the current chunk?)
                        # Note: the current sample position has already been incremented when the sample data was received so
                        # we need to look backwards from the current position to determine if marker falls in chunk window.
                        for marker_id in range(layer.marker_count):
                            if layer_sample_pos - layer_chunk_bytes <= g_array_index_uint(layer.markers, marker_id) < layer_sample_pos:
                                # Marker is in window, send notification
                                send_sound_marker_notification(0,
                                                               layer.sound_id,
                                                               layer.sound_loop_set_id,
                                                               track,
                                                               marker_id)

                    layer_track_buffer_pos_offset += layer_chunk_bytes
                    layer_sample_pos += layer_chunk_bytes
                    layer_bytes_remaining -= layer_chunk_bytes

            # Move to next layer
            layer_iterator = layer_iterator.next

        # Advance buffer pointers
        player.sample_pos += current_chunk_bytes
        track_buffer_pos += current_chunk_bytes
        buffer_bytes_remaining -= current_chunk_bytes

        # If player is counting down to stop, decrement remaining samples and see if player is done
        if player.stop_loop_samples_remaining != do_not_stop_loop:
            player.stop_loop_samples_remaining -= current_chunk_bytes

            if player.stop_loop_samples_remaining == 0:
                player.status = player_idle

        # Stop looping and generating samples if we have reached the end of the sound or
        # the player is now idle (due to reaching end of a fade out)
        if player.status == player_idle or player.sample_pos >= player.length:
            break

    return track_buffer_pos


cdef inline SoundLoopLayerSettings *_create_sound_loop_layer_settings() nogil:
    """
    Creates a new sound loop layer settings struct.
    :return: A pointer to the new settings struct.
    """
    return <SoundLoopLayerSettings*>g_slice_alloc0(sizeof(SoundLoopLayerSettings))
